"""
General helper functions aiding in the boostrapping of this library.
"""

import argparse
import logging
import os
import sys
import time
import weakref
from pathlib import Path
from typing import Any, Optional

import moderngl

from moderngl_window.conf import settings
from moderngl_window.context.base import BaseWindow, WindowConfig
from moderngl_window.timers.clock import Timer
from moderngl_window.utils.keymaps import AZERTY, QWERTY, KeyMap, KeyMapFactory  # noqa
from moderngl_window.utils.module_loading import import_string

__version__ = "3.1.1"

IGNORE_DIRS = [
    "__pycache__",
    "base",
]

# Add new windows classes here to be recognized by the command line option --window
WINDOW_CLASSES = [
    "glfw",
    "headless",
    "pygame2",
    "pyglet",
    "pyqt5",
    "pyside2",
    "sdl2",
    "tk",
]

OPTIONS_TRUE = ["yes", "on", "true", "t", "y", "1"]
OPTIONS_FALSE = ["no", "off", "false", "f", "n", "0"]
OPTIONS_ALL = OPTIONS_TRUE + OPTIONS_FALSE

# Quick and dirty debug logging setup by default
# See: https://docs.python.org/3/howto/logging.html#logging-advanced-tutorial
logger = logging.getLogger(__name__)


def setup_basic_logging(level: int) -> None:
    """Set up basic logging

    Args:
        level (int): The log level
    """
    if level is None:
        return

    # Do not add a new handler if we already have one
    if not logger.handlers:
        logger.propagate = False
        logger.setLevel(level)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        ch.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
        logger.addHandler(ch)


class ContextRefs:
    """Namespace for window/context references"""

    WINDOW: Optional[BaseWindow] = None
    CONTEXT: Optional[moderngl.Context] = None


def activate_context(
    window: Optional[BaseWindow] = None, ctx: Optional[moderngl.Context] = None
) -> None:
    """
    Register the active window and context.
    If only a window is supplied the context is taken from the window.
    Only a context can also be passed in.

    Keyword Args:
        window (window): The window to activate
        ctx (moderngl.Context): The moderngl context to activate
    """
    ContextRefs.WINDOW = window
    ContextRefs.CONTEXT = ctx
    if ctx is None:
        ContextRefs.CONTEXT = window.ctx


def window() -> BaseWindow:
    """Obtain the active window"""
    if ContextRefs.WINDOW:
        return ContextRefs.WINDOW

    raise ValueError("No active window and context. Call activate_window.")


def ctx() -> moderngl.Context:
    """Obtain the active context"""
    if ContextRefs.CONTEXT:
        return ContextRefs.CONTEXT

    raise ValueError("No active window and context. Call activate_window.")


def get_window_cls(window: str = "") -> type[BaseWindow]:
    """
    Attempt to obtain a window class using the full dotted
    python path. This can be used to import custom or modified
    window classes.

    Args:
        window (str): Name of the window

    Returns:
        A reference to the requested window class. Raises exception if not found.
    """
    logger.info("Attempting to load window class: %s", window)
    win = import_string(window)

    # assert issubclass(
    #     win, BaseWindow
    # ), f"{win} is not derived from moderngl_window.context.base.BaseWindow"
    return win


def get_local_window_cls(window: Optional[str] = None) -> type[BaseWindow]:
    """
    Attempt to obtain a window class in the moderngl_window package
    using short window names such as ``pyglet`` or ``glfw``.

    Args:
        window (str): Name of the window

    Returns:
        A reference to the requested window class. Raises exception if not found.
    """
    window = os.environ.get("MODERNGL_WINDOW") or window
    if window is None:
        window = "pyglet"

    return get_window_cls("moderngl_window.context.{}.Window".format(window))


def find_window_classes() -> list[str]:
    """
    Find available window packages
    Returns:
        A list of available window packages
    """
    # In some environments we cannot rely on introspection
    # and instead return a hardcoded list
    try:
        return [
            path.parts[-1]
            for path in Path(__file__).parent.joinpath("context").iterdir()
            if path.is_dir() and path.parts[-1] not in IGNORE_DIRS
        ]
    except Exception:
        return WINDOW_CLASSES


def create_window_from_settings() -> BaseWindow:
    """
    Creates a window using configured values in :py:attr:`moderngl_window.conf.Settings.WINDOW`.
    This will also activate the window/context.

    Returns:
        The Window instance
    """
    window_cls = import_string(settings.WINDOW["class"])
    window = window_cls(**settings.WINDOW)

    assert isinstance(
        window, BaseWindow
    ), f"{type(window)} is not derived from moderngl_window.context.base.BaseWindow"
    activate_context(window=window)
    return window


# --- The simple window config system ---


def run_window_config(
    config_cls: type[WindowConfig], timer: Optional[Timer] = None, args: Any = None
) -> None:
    """
    Run an WindowConfig entering a blocking main loop

    Args:
        config_cls: The WindowConfig class to render
    Keyword Args:
        timer: A custom timer instance
        args: Override sys.args
    """
    config = create_window_config_instance(config_cls, timer=timer, args=args)
    run_window_config_instance(config)


def create_window_config_instance(
    config_cls: type[WindowConfig], timer: Optional[Timer] = None, args: Any = None
) -> WindowConfig:
    """
    Create and initialize a instance of a WindowConfig class.
    Quite a bit of boilerplate is required to create a WindowConfig instance
    and this function aims to simplify that.

    Args:
        window_config: The WindowConfig class to create an instance of
    Keyword Args:
        kwargs: Arguments to pass to the WindowConfig constructor
    Returns:
        An instance of the WindowConfig class
    """
    setup_basic_logging(config_cls.log_level)
    parser = create_parser()
    config_cls.add_arguments(parser)
    values = parse_args(args=args, parser=parser)
    config_cls.argv = values
    window_cls = get_local_window_cls(values.window)

    # Calculate window size
    size = values.size or config_cls.window_size
    size = int(size[0] * values.size_mult), int(size[1] * values.size_mult)

    # Resolve cursor
    show_cursor = values.cursor
    if show_cursor is None:
        show_cursor = config_cls.cursor

    window = window_cls(
        title=config_cls.title,
        size=size,
        fullscreen=config_cls.fullscreen or values.fullscreen,
        resizable=(values.resizable if values.resizable is not None else config_cls.resizable),
        visible=config_cls.visible,
        gl_version=config_cls.gl_version,
        aspect_ratio=config_cls.aspect_ratio,
        vsync=values.vsync if values.vsync is not None else config_cls.vsync,
        samples=values.samples if values.samples is not None else config_cls.samples,
        cursor=show_cursor if show_cursor is not None else True,
        backend=values.backend,
        context_creation_func=config_cls.init_mgl_context,
    )
    window.print_context_info()
    activate_context(window=window)
    if timer is None:
        timer = Timer()
    config = config_cls(ctx=window.ctx, wnd=window, timer=timer)
    # Avoid the event assigning in the property setter for now
    # We want the even assigning to happen in WindowConfig.__init__
    # so users are free to assign them in their own __init__.
    window._config = weakref.ref(config)

    # Swap buffers once before staring the main loop.
    # This can trigger additional resize events reporting
    # a more accurate buffer size
    window.swap_buffers()
    window.set_default_viewport()
    return config


def run_window_config_instance(config: WindowConfig) -> None:
    """
    Run an WindowConfig instance entering a blocking main loop.

    Args:
        window_config: The WindowConfig instance
    """
    window = config.wnd
    timer = config.timer

    timer.start()

    while not window.is_closing:
        current_time, delta = timer.next_frame()

        # Framerate  limit for hidden windows
        if not window.visible and config.hidden_window_framerate_limit > 0:
            expected_delta_time = 1.0 / config.hidden_window_framerate_limit
            sleep_time = expected_delta_time - delta
            if sleep_time > 0:
                time.sleep(sleep_time)

        if config.clear_color is not None:
            window.clear(*config.clear_color)

        # Always bind the window framebuffer before calling render
        window.use()

        window.render(current_time, delta)

        if not window.is_closing:
            window.swap_buffers()

    _, duration = timer.stop()
    window.destroy()
    if duration > 0:
        logger.info("Duration: {0:.2f}s @ {1:.2f} FPS".format(duration, timer.fps_average))


def create_parser() -> argparse.ArgumentParser:
    """Create an argparser parsing the standard arguments for WindowConfig"""
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-wnd",
        "--window",
        choices=find_window_classes(),
        help="Name for the window type to use",
    )
    parser.add_argument(
        "-fs",
        "--fullscreen",
        action="store_true",
        help="Open the window in fullscreen mode",
    )
    parser.add_argument(
        "-vs",
        "--vsync",
        type=valid_bool,
        help="Enable or disable vsync",
    )
    parser.add_argument(
        "-r",
        "--resizable",
        type=valid_bool,
        default=None,
        help="Enable/disable window resize",
    )
    parser.add_argument(
        "-hd",
        "--hidden",
        type=valid_bool,
        default=False,
        help="Start the window in hidden mode",
    )
    parser.add_argument(
        "-s",
        "--samples",
        type=int,
        help="Specify the desired number of samples to use for multisampling",
    )
    parser.add_argument(
        "-c",
        "--cursor",
        type=valid_bool,
        help="Enable or disable displaying the mouse cursor",
    )
    parser.add_argument(
        "--size",
        type=valid_window_size,
        help="Window size",
    )
    parser.add_argument(
        "--size_mult",
        type=valid_window_size_multiplier,
        default=1.0,
        help="Multiplier for the window size making it easy scale the window",
    )
    parser.add_argument(
        "--backend",
        help="Specify context backend. This is mostly used to enable EGL in headless mode",
    )
    return parser


def parse_args(
    args: Optional[Any] = None, parser: Optional[argparse.ArgumentParser] = None
) -> argparse.Namespace:
    """Parse arguments from sys.argv

    Passing in your own argparser can be user to extend the parser.

    Keyword Args:
        args: override for sys.argv
        parser: Supply your own argparser instance
    """
    parser = parser or create_parser()
    return parser.parse_args(args or sys.argv[1:])


# --- Validators ---


def valid_bool(value: Optional[str]) -> Optional[bool]:
    """Validator for bool values"""
    if value is None:
        return None

    value = value.lower()
    if value in OPTIONS_TRUE:
        return True

    if value in OPTIONS_FALSE:
        return False

    raise argparse.ArgumentTypeError(f"Boolean value expected. Options: {OPTIONS_ALL}")


def valid_window_size(value: str) -> tuple[int, int]:
    """
    Validator for window size parameter.

    Valid format is "[int]x[int]". For example "1920x1080".
    """
    try:
        width, height = value.split("x")
        return int(width), int(height)
    except ValueError:
        pass

    raise argparse.ArgumentTypeError(
        "Valid size format: int]x[int]. Example '1920x1080'",
    )


def valid_window_size_multiplier(value: str) -> float:
    """Validates window size multiplier

    Must be an integer or float greater than 0
    """
    try:
        val = float(value)
        if val > 0:
            return val
    except ValueError:
        pass

    raise argparse.ArgumentTypeError(
        "Must be a positive int or float",
    )



from typing import Optional

import moderngl
import numpy

from moderngl_window.geometry import AttributeNames
from moderngl_window.opengl.vao import VAO


def bbox(
    size: tuple[float, float, float] = (1.0, 1.0, 1.0),
    name: Optional[str] = None,
    attr_names: type[AttributeNames] = AttributeNames,
) -> VAO:
    """
    Generates a bounding box with (0.0, 0.0, 0.0) as the center.
    This is simply a box with ``LINE_STRIP`` as draw mode.

    Keyword Args:
        size (tuple): x, y, z size of the box
        name (str): Optional name for the VAO
        attr_names (AttributeNames): Attribute names
    Returns:
        A :py:class:`moderngl_window.opengl.vao.VAO` instance
    """
    width, height, depth = size[0] / 2.0, size[1] / 2.0, size[2] / 2.0
    # fmt: off
    pos = numpy.array([
        width, -height, depth,
        width, height, depth,
        -width, -height, depth,
        width, height, depth,
        -width, height, depth,
        -width, -height, depth,
        width, -height, -depth,
        width, height, -depth,
        width, -height, depth,
        width, height, -depth,
        width, height, depth,
        width, -height, depth,
        width, -height, -depth,
        width, -height, depth,
        -width, -height, depth,
        width, -height, -depth,
        -width, -height, depth,
        -width, -height, -depth,
        -width, -height, depth,
        -width, height, depth,
        -width, height, -depth,
        -width, -height, depth,
        -width, height, -depth,
        -width, -height, -depth,
        width, height, -depth,
        width, -height, -depth,
        -width, -height, -depth,
        width, height, -depth,
        -width, -height, -depth,
        -width, height, -depth,
        width, height, -depth,
        -width, height, -depth,
        width, height, depth,
        -width, height, -depth,
        -width, height, depth,
        width, height, depth,
    ], dtype=numpy.float32)
    # fmt: on

    vao = VAO(name or "geometry:cube", mode=moderngl.LINE_STRIP)
    vao.buffer(pos, "3f", [attr_names.POSITION])

    return vao



"""
Loads two gif files into texture arrays and renders them into an offscreen buffer
that is then displayed on the screen using nearest neighbor filtering.

Possible improvements:
- Make the example configurable to load any gif file(s)
- Load the raw byte data for each frame and instead use a palette texture to
  reduce the memory footprint
- Take the gif transparency key into account?
"""

from pathlib import Path

import glm
import moderngl

import moderngl_window as mglw
from moderngl_window import geometry


class AnimatedGif(mglw.WindowConfig):
    title = "Animated Sprite"
    resource_dir = (Path(__file__) / "../../resources").resolve()
    aspect_ratio = 320 / 256
    window_size = 320 * 3, 256 * 3

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.buffer_size = 320, 256
        # Textures
        self.background_texture = self.load_texture_array("textures/animated_sprites/giphy.gif")
        self.background_texture.repeat_x = False
        self.background_texture.repeat_y = False
        self.caveman_texture = self.load_texture_array(
            "textures/animated_sprites/player_2.gif",
            layers=35,  # Number of frames in the gif
        )
        self.caveman_texture.repeat_x = False
        self.caveman_texture.repeat_y = False
        self.caveman_texture.filter = moderngl.NEAREST, moderngl.NEAREST

        # Geometry
        # One pixel quad 0, 0 -> 1.0, 1.0
        self.sprite_geometry = geometry.quad_2d(size=(1.0, 1.0), pos=(0.5, 0.5))
        self.quad_fs = geometry.quad_fs()

        # Programs
        self.sprite_program = self.load_program("programs/animated_sprites/sprite_array.glsl")
        self.texture_program = self.load_program("programs/texture.glsl")

        # Offscreen buffer
        self.offscreen_texture = self.ctx.texture(self.buffer_size, 4)
        self.offscreen_texture.filter = moderngl.NEAREST, moderngl.NEAREST
        self.offscreen = self.ctx.framebuffer(color_attachments=[self.offscreen_texture])

        self.projection = glm.ortho(0, 320, 0, 256, -1.0, 1.0)
        self.sprite_program["projection"].write(self.projection)

    def on_render(self, time, frame_time):
        # Render sprite of offscreen
        self.offscreen.use()
        self.ctx.clear(0.5, 0.5, 0.5, 0.0)

        self.render_sprite(
            self.background_texture,
            frame=int(time * 15) % self.background_texture.layers,
        )
        self.render_sprite(
            self.caveman_texture,
            frame=int(time * 15) % self.caveman_texture.layers,
            blend=True,
            position=(260, 20),
        )

        # Display offscreen
        self.ctx.screen.use()
        self.offscreen_texture.use(location=0)
        self.quad_fs.render(self.texture_program)

    def render_sprite(self, texture: moderngl.TextureArray, blend=False, frame=0, position=(0, 0)):
        if blend:
            self.ctx.enable(moderngl.BLEND)

        texture.use(location=0)
        self.sprite_program["layer_id"] = frame
        self.sprite_program["position"] = position
        self.sprite_geometry.render(self.sprite_program)

        if blend:
            self.ctx.disable(moderngl.BLEND)


if __name__ == "__main__":
    AnimatedGif.run()



import math
import random

import moderngl_window
from moderngl_window.conf import settings
from moderngl_window.timers.clock import Timer


class CustomSetup:
    """
    Custom setup using a class.
    We create the window, main loop and register events.
    """

    def __init__(self):
        # Configure to use pyglet window
        settings.WINDOW["class"] = "moderngl_window.context.pyglet.Window"
        self.wnd = moderngl_window.create_window_from_settings()
        self.ctx = self.wnd.ctx

        # register event methods
        self.wnd.resize_func = self.on_resize
        self.wnd.iconify_func = self.on_iconify
        self.wnd.key_event_func = self.on_key_event
        self.wnd.mouse_position_event_func = self.on_mouse_position_event
        self.wnd.mouse_drag_event_func = self.on_mouse_drag_event
        self.wnd.mouse_scroll_event_func = self.on_mouse_scroll_event
        self.wnd.mouse_press_event_func = self.on_mouse_press_event
        self.wnd.mouse_release_event_func = self.on_mouse_release_event
        self.wnd.unicode_char_entered_func = self.on_unicode_char_entered
        self.wnd.close_func = self.on_close

    def on_render(self, time, frame_time):
        self.ctx.clear(
            (math.sin(time) + 1.0) / 2,
            (math.sin(time + 2) + 1.0) / 2,
            (math.sin(time + 3) + 1.0) / 2,
        )

    def run(self):
        timer = Timer()
        timer.start()

        while not self.wnd.is_closing:
            self.wnd.clear()
            time, frame_time = timer.next_frame()
            self.on_render(time, frame_time)
            self.wnd.swap_buffers()

        self.wnd.destroy()

    def on_resize(self, width: int, height: int):
        print("Window was resized. buffer size is {} x {}".format(width, height))

    def on_iconify(self, iconify: bool):
        """Window hide/minimize and restore"""
        print("Window was iconified:", iconify)

    def on_key_event(self, key, action, modifiers):
        keys = self.wnd.keys

        # Key presses
        if action == keys.ACTION_PRESS:
            if key == keys.SPACE:
                print("SPACE key was pressed")

            # Using modifiers (shift and ctrl)

            if key == keys.Z and modifiers.shift:
                print("Shift + Z was pressed")

            if key == keys.Z and modifiers.ctrl:
                print("ctrl + Z was pressed")

        # Key releases
        elif action == self.wnd.keys.ACTION_RELEASE:
            if key == keys.SPACE:
                print("SPACE key was released")

        # Move the window around with AWSD
        if action == keys.ACTION_PRESS:
            if key == keys.A:
                self.wnd.position = self.wnd.position[0] - 10, self.wnd.position[1]
            if key == keys.D:
                self.wnd.position = self.wnd.position[0] + 10, self.wnd.position[1]
            if key == keys.W:
                self.wnd.position = self.wnd.position[0], self.wnd.position[1] - 10
            if key == keys.S:
                self.wnd.position = self.wnd.position[0], self.wnd.position[1] + 10

            # toggle cursor
            if key == keys.C:
                self.wnd.cursor = not self.wnd.cursor

            # Shuffle window tittle
            if key == keys.T:
                title = list(self.wnd.title)
                random.shuffle(title)
                self.wnd.title = "".join(title)

            # Toggle mouse exclusivity
            if key == keys.M:
                self.wnd.mouse_exclusivity = not self.wnd.mouse_exclusivity

    def on_mouse_position_event(self, x, y, dx, dy):
        print("Mouse position pos={} {} delta={} {}".format(x, y, dx, dy))

    def on_mouse_drag_event(self, x, y, dx, dy):
        print("Mouse drag pos={} {} delta={} {}".format(x, y, dx, dy))

    def on_mouse_scroll_event(self, x_offset, y_offset):
        print("mouse_scroll_event", x_offset, y_offset)

    def on_mouse_press_event(self, x, y, button):
        print("Mouse button {} pressed at {}, {}".format(button, x, y))
        print("Mouse states:", self.wnd.mouse_states)

    def on_mouse_release_event(self, x: int, y: int, button: int):
        print("Mouse button {} released at {}, {}".format(button, x, y))
        print("Mouse states:", self.wnd.mouse_states)

    def on_unicode_char_entered(self, char):
        print("unicode_char_entered:", char)

    def on_close(self):
        print("Window was closed")


if __name__ == "__main__":
    CustomSetup().run()



"""
Wrapper for a loaded mesh / vao with properties
"""

from __future__ import annotations
from typing import Optional

import glm
import moderngl

from moderngl_window.opengl.vao import VAO

from .camera import Camera
from .mesh import Mesh


class Node:
    """A generic scene node containing a mesh or camera
    and/or a container for other nodes. Nodes and their children
    represents the scene tree.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        camera: Optional[Camera] = None,
        mesh: Optional[Mesh] = None,
        matrix: Optional[glm.mat4] = None,
    ):
        """Create a node.

        Keyword Args:
            name: Name of the node
            camera: Camera to store in the node
            mesh: Mesh to store in the node
            matrix: The node's matrix
        """
        self._name = name
        self._camera = camera
        self._mesh = mesh
        # Local node matrix
        self._matrix = matrix
        # Global matrix
        self._matrix_global = glm.mat4(1.0)

        self._children: list["Node"] = []

    @property
    def name(self) -> Optional[str]:
        """str: Get or set the node name"""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def mesh(self) -> Optional[Mesh]:
        """:py:class:`~moderngl_window.scene.Mesh`: The mesh if present"""
        return self._mesh

    @mesh.setter
    def mesh(self, value: Mesh) -> None:
        self._mesh = value

    @property
    def camera(self) -> Optional[Camera]:
        """:py:class:`~moderngl_window.scene.Camera`: The camera if present"""
        return self._camera

    @camera.setter
    def camera(self, value: Camera) -> None:
        self._camera = value

    @property
    def matrix(self) -> Optional[glm.mat4]:
        """glm.mat4x4: Note matrix (local)"""
        return self._matrix

    @matrix.setter
    def matrix(self, value: glm.mat4) -> None:
        self._matrix = value

    @property
    def matrix_global(self) -> Optional[glm.mat4]:
        """glm.matx4: The global node matrix containing transformations from parent nodes"""
        return self._matrix_global

    @matrix_global.setter
    def matrix_global(self, value: glm.mat4) -> None:
        self._matrix_global = value

    @property
    def children(self) -> list["Node"]:
        """list: List of children"""
        return self._children

    def add_child(self, node: "Node") -> None:
        """Add a child to this node

        Args:
            node (Node): Node to add as a child
        """
        self._children.append(node)

    def draw(
        self,
        projection_matrix: glm.mat4,
        camera_matrix: glm.mat4,
        time: float = 0.0,
    ) -> None:
        """Draw node and children.

        Keyword Args:
            projection_matrix: projection matrix
            camera_matrix: camera_matrix
            time: The current time
        """
        if self._mesh:
            self._mesh.draw(
                projection_matrix=projection_matrix,
                model_matrix=self._matrix_global,
                camera_matrix=camera_matrix,
                time=time,
            )

        for child in self._children:
            child.draw(
                projection_matrix=projection_matrix,
                camera_matrix=camera_matrix,
                time=time,
            )

    def draw_bbox(
        self,
        projection_matrix: Optional[glm.mat4],
        camera_matrix: Optional[glm.mat4],
        program: moderngl.Program,
        vao: VAO,
    ) -> None:
        """Draw bounding box around the node and children.

        Keyword Args:
            projection_matrix: projection matrix
            camera_matrix: camera_matrix
            program (moderngl.Program): The program to render the bbox
            vao: The vertex array representing the bounding box
        """
        if self._mesh:
            assert (
                projection_matrix is not None
            ), "Can not draw bbox, the projection matrix is empty"
            assert self._matrix_global is not None, "Can not draw bbox, the global matrix is empty"
            assert camera_matrix is not None, "Can not draw bbox, the camera matrix is empty"
            self._mesh.draw_bbox(
                projection_matrix, self._matrix_global, camera_matrix, program, vao
            )

        for child in self.children:
            child.draw_bbox(projection_matrix, camera_matrix, program, vao)

    def draw_wireframe(
        self,
        projection_matrix: Optional[glm.mat4],
        camera_matrix: Optional[glm.mat4],
        program: moderngl.Program,
    ) -> None:
        """Render the node as wireframe.

        Keyword Args:
            projection_matrix (bytes): projection matrix
            camera_matrix (bytes): camera_matrix
            program (moderngl.Program): The program to render wireframe
        """
        if self._mesh:
            assert (
                projection_matrix is not None
            ), "Can not draw bbox, the projection matrix is empty"
            assert self._matrix_global is not None, "Can not draw bbox, the global matrix is empty"
            self._mesh.draw_wireframe(projection_matrix, self._matrix_global, program)

        for child in self.children:
            child.draw_wireframe(projection_matrix, self._matrix_global, program)

    def calc_global_bbox(
        self, view_matrix: glm.mat4, bbox_min: glm.vec3 | None, bbox_max: glm.vec3 | None
    ) -> tuple[glm.vec3, glm.vec3]:
        """Recursive calculation of scene bbox.

        Keyword Args:
            view_matrix (numpy.ndarray): view matrix
            bbox_min: min bbox values
            bbox_max: max bbox values
        """
        if self._matrix is not None:
            view_matrix = self._matrix * view_matrix

        if self._mesh:
            bbox_min, bbox_max = self._mesh.calc_global_bbox(view_matrix, bbox_min, bbox_max)

        for child in self._children:
            bbox_min, bbox_max = child.calc_global_bbox(view_matrix, bbox_min, bbox_max)

        return bbox_min, bbox_max

    def calc_model_mat(self, parent_matrix: glm.mat4) -> None:
        """Calculate the model matrix related to all parents.

        Args:
            parent_matrix: Matrix for parent node
        """
        if self._matrix is not None:
            self._matrix_global = parent_matrix * self._matrix

            for child in self._children:
                child.calc_model_mat(self._matrix_global)
        else:
            self._matrix_global = parent_matrix

            for child in self._children:
                child.calc_model_mat(parent_matrix)

    def __repr__(self) -> str:
        return "<Node name={}>".format(self.name)



from pathlib import Path

import glm
import moderngl

from imgui_bundle import imgui

import moderngl_window as mglw
from moderngl_window import geometry
from moderngl_window.integrations.imgui_bundle import ModernglWindowRenderer


class WindowEvents(mglw.WindowConfig):
    gl_version = (3, 3)
    title = "imgui Integration"
    resource_dir = (Path(__file__).parent / "../examples/resources").resolve()
    aspect_ratio = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        imgui.create_context()
        self.wnd.ctx.error
        self.imgui = ModernglWindowRenderer(self.wnd)

        self.cube = geometry.cube(size=(2, 2, 2))
        self.prog = self.load_program("programs/cube_simple.glsl")
        self.prog["color"].value = (1.0, 1.0, 1.0, 1.0)
        self.prog["m_camera"].write(glm.mat4())
        self.prog["m_proj"].write(glm.perspective(glm.radians(75), 1.0, 1, 100))

        self.fbo = self.ctx.framebuffer(
            color_attachments=self.ctx.texture((512, 512), 4),
            depth_attachment=self.ctx.depth_texture((512, 512)),
        )
        # Ensure imgui knows about this texture
        # This is the color layer in the framebuffer
        self.imgui.register_texture(self.fbo.color_attachments[0])

    def on_render(self, time: float, frametime: float):
        # Rotate/move cube
        rotation = glm.mat4(glm.quat(glm.vec3(time, time, time)))
        translation = glm.translate(glm.vec3(0.0, 0.0, -3.5))
        model = translation * rotation

        # Render cube to offscreen texture / fbo
        self.fbo.use()
        self.fbo.clear()
        self.ctx.enable(moderngl.DEPTH_TEST | moderngl.CULL_FACE)
        self.prog["m_model"].write(model)
        self.cube.render(self.prog)

        # Render UI to screen
        self.wnd.use()
        self.render_ui()

    def render_ui(self):
        """Render the UI"""
        imgui.new_frame()
        if imgui.begin_main_menu_bar():
            if imgui.begin_menu("File", True):
                clicked_quit, selected_quit = imgui.menu_item("Quit", "Cmd+Q", False, True)

                if clicked_quit:
                    exit(1)

                imgui.end_menu()
            imgui.end_main_menu_bar()

        imgui.show_demo_window()

        imgui.begin("Custom window", True)
        imgui.text("Bar")
        imgui.text_colored(imgui.ImVec4(0.2, 1.0, 0.0, 1.0), "Eggs")
        imgui.end()

        # Create window with the framebuffer image
        imgui.begin("Custom window with Image", True)
        # Create an image control by passing in the OpenGL texture ID (glo)
        # and pass in the image size as well.
        # The texture needs to he registered using register_texture for this to work
        imgui.image(self.fbo.color_attachments[0].glo, self.fbo.size)
        imgui.end()

        imgui.render()
        self.imgui.render(imgui.get_draw_data())

    def on_resize(self, width: int, height: int):
        self.imgui.resize(width, height)

    def on_key_event(self, key, action, modifiers):
        self.imgui.key_event(key, action, modifiers)

    def on_mouse_position_event(self, x, y, dx, dy):
        self.imgui.mouse_position_event(x, y, dx, dy)

    def on_mouse_drag_event(self, x, y, dx, dy):
        self.imgui.mouse_drag_event(x, y, dx, dy)

    def on_mouse_scroll_event(self, x_offset, y_offset):
        self.imgui.mouse_scroll_event(x_offset, y_offset)

    def on_mouse_press_event(self, x, y, button):
        self.imgui.mouse_press_event(x, y, button)

    def on_mouse_release_event(self, x: int, y: int, button: int):
        self.imgui.mouse_release_event(x, y, button)

    def on_unicode_char_entered(self, char):
        self.imgui.unicode_char_entered(char)


if __name__ == "__main__":
    WindowEvents.run()



import platform
from pathlib import Path

import moderngl
import pytest
from headless import HeadlessTestCase

from moderngl_window import resources
from moderngl_window.exceptions import ImproperlyConfigured
from moderngl_window.meta import ProgramDescription
from moderngl_window.opengl.program import ReloadableProgram

resources.register_dir((Path(__file__).parent / 'fixtures' / 'resources').resolve())


class ProgramLoadersTestCase(HeadlessTestCase):
    window_size = (16, 16)
    aspect_ratio = 1.0

    def test_single(self):
        """Load a single file glsl program"""
        program = resources.programs.load(ProgramDescription(path='programs/white.glsl'))
        self.assertIsInstance(program, moderngl.Program)
        # Ensure attribute is present
        program['in_position']
        self.assertIsInstance(program.extra.get('meta'), ProgramDescription)

    def test_single_feedback(self):
        """Load transform feedback shader"""
        program = resources.programs.load(ProgramDescription(path='programs/feedback.glsl'))
        self.assertIsInstance(program, moderngl.Program)

    def test_single_geometry(self):
        """Load single glsl file with gemotry shader"""
        program = resources.programs.load(ProgramDescription(path='programs/billboards/billboards.glsl'))
        self.assertIsInstance(program, moderngl.Program)

    def test_single_tesselation(self):
        """Load single glsl file with tessellation"""
        program = resources.programs.load(ProgramDescription(path='programs/terrain/terrain.glsl'))
        self.assertIsInstance(program, moderngl.Program)

    def test_single_reloadable(self):
        """Load a single file glsl program as reloadable"""
        program = resources.programs.load(ProgramDescription(path='programs/white.glsl', reloadable=True))
        self.assertIsInstance(program, ReloadableProgram)
        # Ensure attribute is present
        program['in_position']
        self.assertIsInstance(program.extra.get('meta'), ProgramDescription)
        self.assertEqual(program.name, 'programs/white.glsl')
        self.assertIsInstance(program.ctx, moderngl.Context)
        self.assertIsNotNone(program.get('in_position', None))
        self.assertIsNotNone(program.mglo)
        self.assertGreater(program.glo, 0)
        program.subroutines
        program.geometry_input
        program.geometry_output
        program.geometry_vertices
        program._members
        repr(program)

    def test_separate_geometry(self):
        program = resources.programs.load(ProgramDescription(
            vertex_shader="programs/billboards/billboards_vs.glsl",
            geometry_shader="programs/billboards/billboards_gs.glsl",
            fragment_shader="programs/billboards/billboards_fs.glsl",
        ))
        self.assertIsInstance(program, moderngl.Program)

    def test_separate_tesselation(self):
        """Load a more complex tesselation program from separate files"""
        program = resources.programs.load(ProgramDescription(
            vertex_shader="programs/terrain/terrain_vs.glsl",
            tess_control_shader="programs/terrain/terrain_tc.glsl",
            tess_evaluation_shader="programs/terrain/terrain_te.glsl",
            fragment_shader="programs/terrain/terrain_fs.glsl",
        ))
        self.assertIsInstance(program, moderngl.Program)

    def test_separate_relodable(self):
        """Load a more complex tesselation program from separate files"""
        program = resources.programs.load(ProgramDescription(
            vertex_shader="programs/terrain/terrain_vs.glsl",
            tess_control_shader="programs/terrain/terrain_tc.glsl",
            tess_evaluation_shader="programs/terrain/terrain_te.glsl",
            fragment_shader="programs/terrain/terrain_fs.glsl",
            reloadable=True,
        ))
        self.assertIsInstance(program, ReloadableProgram)

    def test_single_not_found(self):
        """Ensure ImproperlyConfigured is raised when shaders are not found"""
        with self.assertRaises(ImproperlyConfigured):
            resources.programs.load(ProgramDescription(path='programs/nonexist.glsl'))

    def test_separate_not_found(self):
        """Ensure ImproperlyConfigured is raised when shaders are not found"""
        with self.assertRaises(ImproperlyConfigured):
            resources.programs.load(ProgramDescription(
                vertex_shader="programs/notfound_vs.glsl",
                geometry_shader="programs/notfound_geo.glsl",
                tess_control_shader="programs/notfound_tc.glsl",
                tess_evaluation_shader="programs/notfound_te.glsl",
                fragment_shader="programs/notfound_fs.glsl",
            ))

    @pytest.mark.skipif(platform.system() == 'Darwin', reason="Compute shader not possible on OSX")
    def test_compute_shader(self):
        """Attempt loading a compute shader"""
        path = 'programs/compute.glsl'
        descr = ProgramDescription(compute_shader=path)
        self.assertEqual(descr.compute_shader, path)
        program = resources.programs.load(descr)
        self.assertIsInstance(program, moderngl.ComputeShader)

    def test_include(self):
        program = resources.programs.load(ProgramDescription(path='programs/include_test.glsl'))
        self.assertIsInstance(program, moderngl.Program)

    def test_varyings_override(self):
        """Specify varyings during program creation"""
        path = 'programs/varyings.glsl'
        descr = ProgramDescription(vertex_shader=path, varyings=["value_1",  "value_2"])
        program = resources.programs.load(descr)



import logging
import os
from datetime import datetime
from typing import Optional, Union

import moderngl
from PIL import Image

from moderngl_window.conf import settings

logger = logging.getLogger(__name__)

TEXTURE_MODES = [None, "L", None, "RGB", "RGBA"]


def create(
    source: Union[moderngl.Framebuffer, moderngl.Texture],
    file_format: str = "png",
    name: Optional[str] = None,
    mode: str = "RGB",
    alignment: int = 1,
) -> None:
    """
    Create a screenshot from a ``moderngl.Framebuffer`` or ``moderngl.Texture``.
    The screenshot will be written to :py:attr:`~moderngl_window.conf.Settings.SCREENSHOT_PATH`
    if set or ``cwd`` or an absolute path can be used.

    Args:
        source: The framebuffer or texture to screenshot
        file_format (str): formats supported by PIL (png, jpeg etc)
        name (str): Optional file name with relative or absolute path
        mode (str): Components/mode to use
        alignment (int): Buffer alignment
    """
    dest = ""
    if settings.SCREENSHOT_PATH:
        if not os.path.exists(str(settings.SCREENSHOT_PATH)):
            logger.debug("SCREENSHOT_PATH does not exist. creating: %s", settings.SCREENSHOT_PATH)
            os.makedirs(str(settings.SCREENSHOT_PATH))
        dest = settings.SCREENSHOT_PATH
    else:
        logger.info("SCREENSHOT_PATH not defined in settings. Using cwd as fallback.")

    if not name:
        name = "{}.{}".format(datetime.now().strftime("%Y-%m-%d-%H-%M-%S-%f"), file_format)

    logger.debug(
        "Creating screenshot: source=%s file_format=%s name=%s mode=%s alignment=%s",
        source,
        file_format,
        name,
        mode,
        alignment,
    )

    if isinstance(source, moderngl.Framebuffer):
        image = Image.frombytes(
            mode,
            (
                source.viewport[2] - source.viewport[0],
                source.viewport[3] - source.viewport[1],
            ),
            source.read(viewport=source.viewport, alignment=alignment),
        )
    elif isinstance(source, moderngl.Texture):
        image = Image.frombytes(
            TEXTURE_MODES[source.components], source.size, source.read(alignment=1)
        )
    else:
        raise ValueError("Source needs to be a FrameBuffer or Texture, not a %s", type(source))

    image = image.transpose(Image.Transpose.FLIP_TOP_BOTTOM)
    dest = os.path.join(str(dest), name)
    logger.info("Creating screenshot: %s", dest)
    image.save(dest, format=file_format)



from pathlib import Path
from unittest import mock

from headless import HeadlessTestCase
from utils import settings_context

from moderngl_window import screenshot


class ScreenshotTestCase(HeadlessTestCase):

    @mock.patch('PIL.Image.Image.save', new=mock.MagicMock())
    def test_fbo(self):
        """Create screenshot from fbo"""
        screenshot.create(self.window.fbo)

    @mock.patch('PIL.Image.Image.save', new=mock.MagicMock())
    def test_texture(self):
        """Create screenshot from texture"""
        texture = self.ctx.texture((16, 16), 4)
        screenshot.create(texture)

    @mock.patch('os.makedirs', new=mock.MagicMock())
    def test_with_screenshot_path(self):
        """Create screenshot with SCREENSHOT_PATH defined in settings"""
        with settings_context({'SCREENSHOT_PATH': (Path.cwd() / 'screenshots')}):
            self.test_fbo()

    def test_incorrect_source(self):
        """Attempt to pass invalid source"""
        with self.assertRaises(ValueError):
            screenshot.create("Hello")



from pathlib import Path

import moderngl_window
from moderngl_window import geometry, resources

resources.register_dir((Path(__file__).parent / "resources").resolve())


class QuadFullscreen(moderngl_window.WindowConfig):
    title = "Quad Fullscreen"
    window_size = 1980, 1024
    aspect_ratio = 1980 / 1024

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.quad = geometry.quad_fs()
        self.texture = self.load_texture_2d("textures/python-bg.png")
        self.prog = self.load_program("programs/texture.glsl")

    def on_render(self, time: float, frame_time: float):
        self.ctx.clear()

        self.texture.use(location=0)
        self.prog["texture0"].value = 0
        self.quad.render(self.prog)


if __name__ == "__main__":
    moderngl_window.run_window_config(QuadFullscreen)



import time
from typing import Any, Optional, Union

import glm
from glm import cos, radians, sin

from moderngl_window.context.base import BaseKeys
from moderngl_window.opengl.projection import Projection3D
from moderngl_window.utils.keymaps import QWERTY, KeyMapFactory

# Direction Definitions
RIGHT = 1
LEFT = 2
FORWARD = 3
BACKWARD = 4
UP = 5
DOWN = 6

# Movement Definitions
STILL = 0
POSITIVE = 1
NEGATIVE = 2


class Camera:
    """Simple camera class containing projection.

    .. code:: python

        # create a camera
        camera = Camera(fov=60.0, aspect_ratio=1.0, near=1.0, far=100.0)

        # Get the current camera matrix as numpy array
        print(camera.matrix)

        # Get projection matrix as numpy array
        print(camera.projection.matrix)
    """

    def __init__(
        self, fov: float = 60.0, aspect_ratio: float = 1.0, near: float = 1.0, far: float = 100.0
    ):
        """Initialize camera using a specific projection

        Keyword Args:
            fov (float): Field of view
            aspect_ratio (float): Aspect ratio
            near (float): Near plane
            far (float): Far plane
        """
        self.position = glm.vec3(0.0, 0.0, 0.0)
        # Default camera placement
        self.up = glm.vec3(0.0, 1.0, 0.0)
        self.right = glm.vec3(1.0, 0.0, 0.0)
        self.dir = glm.vec3(0.0, 0.0, -1.0)
        # Yaw and Pitch
        self._yaw = -90.0
        self._pitch = 0.0

        # World up vector
        self._up = glm.vec3(0.0, 1.0, 0.0)

        # Projection
        self._projection = Projection3D(aspect_ratio, fov, near, far)

    @property
    def projection(self) -> Projection3D:
        """:py:class:`~moderngl_window.opengl.projection.Projection3D`: The 3D projection"""
        return self._projection

    def set_position(self, x: float, y: float, z: float) -> None:
        """Set the 3D position of the camera.

        Args:
            x (float): x position
            y (float): y position
            z (float): z position
        """
        self.position = glm.vec3(x, y, z)

    def set_rotation(self, yaw: float, pitch: float) -> None:
        """Set the rotation of the camera.

        Args:
            yaw (float): yaw rotation
            pitch (float): pitch rotation
        """
        self._pitch = pitch
        self._yaw = yaw
        self._update_yaw_and_pitch()

    @property
    def yaw(self) -> float:
        """float: The current yaw angle."""
        return self._yaw

    @yaw.setter
    def yaw(self, value: float) -> None:
        self._yaw = value
        self._update_yaw_and_pitch()

    @property
    def pitch(self) -> float:
        """float: The current pitch angle."""
        return self._pitch

    @pitch.setter
    def pitch(self, value: float) -> None:
        self._pitch = value
        self._update_yaw_and_pitch()

    @property
    def matrix(self) -> glm.mat4:
        """glm.mat4: The current view matrix for the camera"""
        self._update_yaw_and_pitch()
        return self._gl_look_at(self.position, self.position + self.dir, self._up)

    def _update_yaw_and_pitch(self) -> None:
        """Updates the camera vectors based on the current yaw and pitch"""
        front = glm.vec3(0.0, 0.0, 0.0)
        front.x = cos(radians(self.yaw)) * cos(radians(self.pitch))
        front.y = sin(radians(self.pitch))
        front.z = sin(radians(self.yaw)) * cos(radians(self.pitch))

        self.dir = glm.normalize(front)
        self.right = glm.normalize(glm.cross(self.dir, self._up))
        self.up = glm.normalize(glm.cross(self.right, self.dir))

    def look_at(
        self, vec: Optional[glm.vec3] = None, pos: Optional[tuple[float, float, float]] = None
    ) -> glm.mat4:
        """Look at a specific point

        Either ``vec`` or ``pos`` needs to be supplied.

        Keyword Args:
            vec (glm.vec3): position
            pos (tuple/list): list of tuple ``[x, y, x]`` / ``(x, y, x)``
        Returns:
            glm.mat4x4: Camera matrix
        """
        if pos is not None:
            vec = glm.vec3(pos)

        if vec is None:
            raise ValueError("vector or pos must be set")

        return self._gl_look_at(self.position, vec, self._up)

    def _gl_look_at(self, pos: glm.vec3, target: glm.vec3, up: glm.vec3) -> glm.mat4:
        """The standard lookAt method.

        Args:
            pos: current position
            target: target position to look at
            up: direction up
        Returns:
            glm.mat4: The matrix
        """
        z = glm.normalize(pos - target)
        x = glm.normalize(glm.cross(glm.normalize(up), z))
        y = glm.cross(z, x)

        translate = glm.mat4()
        translate[3][0] = -pos.x
        translate[3][1] = -pos.y
        translate[3][2] = -pos.z

        rotate = glm.mat4()
        rotate[0][0] = x[0]  # -- X
        rotate[1][0] = x[1]
        rotate[2][0] = x[2]
        rotate[0][1] = y[0]  # -- Y
        rotate[1][1] = y[1]
        rotate[2][1] = y[2]
        rotate[0][2] = z[0]  # -- Z
        rotate[1][2] = z[1]
        rotate[2][2] = z[2]

        return rotate * translate


class KeyboardCamera(Camera):
    """Camera controlled by mouse and keyboard.
    The class interacts with the key constants in the
    built in window types.

    Creating a keyboard camera:

    .. code:: python

        camera = KeyboardCamera(
            self.wnd.keys,
            fov=75.0,
            aspect_ratio=self.wnd.aspect_ratio,
            near=0.1,
            far=1000.0,
        )

    We can also interact with the belonging
    :py:class:`~moderngl_window.opengl.projection.Projection3D` instance.

    .. code:: python

        # Update aspect ratio
        camera.projection.update(aspect_ratio=1.0)

        # Get projection matrix in bytes (f4)
        camera.projection.tobytes()
    """

    def __init__(
        self,
        keys: BaseKeys,
        keymap: KeyMapFactory = QWERTY,
        fov: float = 60.0,
        aspect_ratio: float = 1.0,
        near: float = 1.0,
        far: float = 100.0,
    ):
        """Initialize the camera

        Args:
            keys (BaseKeys): The key constants for the current window type
        Keyword Args:
            keymap (KeyMapFactory) : The keymap to use. By default QWERTY.
            fov (float): Field of view
            aspect_ratio (float): Aspect ratio
            near (float): near plane
            far (float): far plane
        """
        # Position movement states
        self.keys = keys
        self.keymap = keymap(keys)
        self._xdir = STILL
        self._zdir = STILL
        self._ydir = STILL
        self._last_time = 0.0
        self._last_rot_time = 0.0

        # Velocity in axis units per second
        self._velocity = 10.0
        self._mouse_sensitivity = 0.5

        super().__init__(fov=fov, aspect_ratio=aspect_ratio, near=near, far=far)

    @property
    def mouse_sensitivity(self) -> float:
        """float: Mouse sensitivity (rotation speed).

        This property can also be set::

            camera.mouse_sensitivity = 2.5
        """
        return self._mouse_sensitivity

    @mouse_sensitivity.setter
    def mouse_sensitivity(self, value: float) -> None:
        self._mouse_sensitivity = value

    @property
    def velocity(self) -> float:
        """float: The speed this camera move based on key inputs

        The property can also be modified::

            camera.velocity = 5.0
        """
        return self._velocity

    @velocity.setter
    def velocity(self, value: float) -> None:
        self._velocity = value

    def key_input(self, key: str, action: str, modifiers: Any) -> None:
        """Process key inputs and move camera

        Args:
            key: The key
            action: key action release/press
            modifiers: key modifier states such as ctrl or shit
        """
        # Right
        if key == self.keymap.RIGHT:
            if action == self.keys.ACTION_PRESS:
                self.move_right(True)
            elif action == self.keys.ACTION_RELEASE:
                self.move_right(False)
        # Left
        elif key == self.keymap.LEFT:
            if action == self.keys.ACTION_PRESS:
                self.move_left(True)
            elif action == self.keys.ACTION_RELEASE:
                self.move_left(False)
        # Forward
        elif key == self.keymap.FORWARD:
            if action == self.keys.ACTION_PRESS:
                self.move_forward(True)
            if action == self.keys.ACTION_RELEASE:
                self.move_forward(False)
        # Backwards
        elif key == self.keymap.BACKWARD:
            if action == self.keys.ACTION_PRESS:
                self.move_backward(True)
            if action == self.keys.ACTION_RELEASE:
                self.move_backward(False)

        # DOWN
        elif key == self.keymap.DOWN:
            if action == self.keys.ACTION_PRESS:
                self.move_down(True)
            if action == self.keys.ACTION_RELEASE:
                self.move_down(False)

        # UP
        elif key == self.keymap.UP:
            if action == self.keys.ACTION_PRESS:
                self.move_up(True)
            if action == self.keys.ACTION_RELEASE:
                self.move_up(False)

    def move_left(self, activate: bool) -> None:
        """The camera should be continiously moving to the left.

        Args:
            activate (bool): Activate or deactivate this state
        """
        self.move_state(LEFT, activate)

    def move_right(self, activate: bool) -> None:
        """The camera should be continiously moving to the right.

        Args:
            activate (bool): Activate or deactivate this state
        """
        self.move_state(RIGHT, activate)

    def move_forward(self, activate: bool) -> None:
        """The camera should be continiously moving forward.

        Args:
            activate (bool): Activate or deactivate this state
        """
        self.move_state(FORWARD, activate)

    def move_backward(self, activate: bool) -> None:
        """The camera should be continiously moving backwards.

        Args:
            activate (bool): Activate or deactivate this state
        """
        self.move_state(BACKWARD, activate)

    def move_up(self, activate: bool) -> None:
        """The camera should be continiously moving up.

        Args:
            activate (bool): Activate or deactivate this state
        """
        self.move_state(UP, activate)

    def move_down(self, activate: bool) -> None:
        """The camera should be continiously moving down.

        Args:
            activate (bool): Activate or deactivate this state
        """
        self.move_state(DOWN, activate)

    def move_state(self, direction: int, activate: bool) -> None:
        """Set the camera position move state.

        Args:
            direction: What direction to update
            activate: Start or stop moving in the direction
        """
        if direction == RIGHT:
            self._xdir = POSITIVE if activate else STILL
        elif direction == LEFT:
            self._xdir = NEGATIVE if activate else STILL
        elif direction == FORWARD:
            self._zdir = NEGATIVE if activate else STILL
        elif direction == BACKWARD:
            self._zdir = POSITIVE if activate else STILL
        elif direction == UP:
            self._ydir = POSITIVE if activate else STILL
        elif direction == DOWN:
            self._ydir = NEGATIVE if activate else STILL

    def rot_state(self, dx: float, dy: float) -> None:
        """Update the rotation of the camera.

        This is done by passing in the relative
        mouse movement change on x and y (delta x, delta y).

        In the past this method took the viewport position
        of the mouse. This does not work well when
        mouse exclusivity mode is enabled.

        Args:
            dx: Relative mouse position change on x
            dy: Relative mouse position change on y
        """
        now = time.time()
        delta = now - self._last_rot_time
        self._last_rot_time = now

        # Greatly decrease the chance of camera popping.
        # This can happen when the mouse enters and leaves the window
        # or when getting focus again.
        if delta > 0.1 and max(abs(dx), abs(dy)) > 2:
            return

        dx *= self._mouse_sensitivity
        dy *= self._mouse_sensitivity

        self._yaw -= dx
        self._pitch += dy

        if self.pitch > 85.0:
            self.pitch = 85.0
        if self.pitch < -85.0:
            self.pitch = -85.0

        self._update_yaw_and_pitch()

    @property
    def matrix(self) -> glm.mat4:
        """glm.mat4x4: The current view matrix for the camera"""
        # Use separate time in camera so we can move it when the demo is paused
        now = time.time()
        # If the camera has been inactive for a while, a large time delta
        # can suddenly move the camera far away from the scene
        t = max(now - self._last_time, 0)
        self._last_time = now

        # X Movement
        if self._xdir == POSITIVE:
            self.position += self.right * self._velocity * t
        elif self._xdir == NEGATIVE:
            self.position -= self.right * self._velocity * t

        # Z Movement
        if self._zdir == NEGATIVE:
            self.position += self.dir * self._velocity * t
        elif self._zdir == POSITIVE:
            self.position -= self.dir * self._velocity * t

        # Y Movement
        if self._ydir == POSITIVE:
            self.position += self.up * self._velocity * t
        elif self._ydir == NEGATIVE:
            self.position -= self.up * self._velocity * t

        return self._gl_look_at(self.position, self.position + self.dir, self._up)


class OrbitCamera(Camera):
    """Camera controlled by the mouse to pan around the target.

    The functions :py:function:`~camera.OrbitCamera.rot_state` and
    :py:function:`~camera.OrbitCamera.rot_state` are used to update the rotation and zoom.

    Creating a orbit camera:

    .. code:: python

        camera = OrbitCamera(
            target=(0., 0., 0.),
            radius=2.0
            fov=75.0,
            aspect_ratio=self.wnd.aspect_ratio,
            near=0.1,
            far=1000.0,
        )

    We can also interact with the belonging
    :py:class:`~moderngl_window.opengl.projection.Projection3D` instance.

    .. code:: python

        # Update aspect ratio
        camera.projection.update(aspect_ratio=1.0)

        # Get projection matrix in bytes (f4)
        camera.projection.tobytes()
    """

    def __init__(
        self,
        target: Union[glm.vec3, tuple[float, float, float]] = (0.0, 0.0, 0.0),
        radius: float = 2.0,
        angles: tuple[float, float] = (45.0, -45.0),
        **kwargs: Any,
    ):
        """Initialize the camera

        Keyword Args:
            target (float, float, float): Target point
            radius (float): Radius
            angles (float, float): angle_x and angle_y in degrees
            fov (float): Field of view
            aspect_ratio (float): Aspect ratio
            near (float): near plane
            far (float): far plane
        """
        # values for orbit camera
        self.radius = radius  # radius in base units
        self.angle_x, self.angle_y = angles  # angles in degrees
        self.target = glm.vec3(target)  # camera target in base units
        self.up = glm.vec3(0.0, 1.0, 0.0)  # camera up vector

        self._mouse_sensitivity = 1.0
        self._zoom_sensitivity = 1.0

        super().__init__(**kwargs)

    @property
    def matrix(self) -> glm.mat4:
        """glm.mat4: The current view matrix for the camera"""
        # Compute camera (eye) position, calculated from angles and radius.
        position = (
            cos(radians(self.angle_x)) * sin(radians(self.angle_y)) * self.radius + self.target[0],
            cos(radians(self.angle_y)) * self.radius + self.target[1],
            sin(radians(self.angle_x)) * sin(radians(self.angle_y)) * self.radius + self.target[2],
        )
        self.set_position(*position)
        return glm.lookAt(
            position,
            self.target,  # what to look at
            self.up,  # camera up direction (change for rolling the camera)
        )

    @property
    def angle_x(self) -> float:
        """float: camera angle x in degrees.

        This property can also be set::
            camera.angle_x = 45.
        """
        return self._angle_x

    @angle_x.setter
    def angle_x(self, value: float) -> None:
        """Set camera rotation_x in degrees."""
        self._angle_x = value

    @property
    def angle_y(self) -> float:
        """float: camera angle y in degrees.

        This property can also be set::
            camera.angle_y = 45.
        """
        return self._angle_y

    @angle_y.setter
    def angle_y(self, value: float) -> None:
        """Set camera rotation_y in degrees."""
        self._angle_y = value

    @property
    def mouse_sensitivity(self) -> float:
        """float: Mouse sensitivity (rotation speed).

        This property can also be set::
            camera.mouse_sensitivity = 2.5
        """
        return self._mouse_sensitivity

    @mouse_sensitivity.setter
    def mouse_sensitivity(self, value: float) -> None:
        self._mouse_sensitivity = value

    @property
    def zoom_sensitivity(self) -> float:
        """float: Mousewheel zooming sensitivity (zoom speed).

        This property can also be set::
            camera.zoom_sensitivity = 2.5
        """
        return self._zoom_sensitivity

    @zoom_sensitivity.setter
    def zoom_sensitivity(self, value: float) -> None:
        self._zoom_sensitivity = value

    def rot_state(self, dx: float, dy: float) -> None:
        """Update the rotation of the camera around the target point.

        This is done by passing relative mouse change in the x and y axis (delta x, delta y)

        Args:
            dx: Relative mouse position change on x axis
            dy: Relative mouse position change on y axis
        """
        self.angle_x += dx * self.mouse_sensitivity / 10.0
        self.angle_y += dy * self.mouse_sensitivity / 10.0

        # clamp the y angle to avoid weird rotations
        self.angle_y = max(min(self.angle_y, -5.0), -175.0)

    def zoom_state(self, y_offset: float) -> None:
        # allow zooming in/out
        self.radius -= y_offset * self._zoom_sensitivity
        self.radius = max(1.0, self.radius)



"""
Cubes of different vertex formats.
These formats are unique for wavefront obj files.
"""

from pathlib import Path

import glm
import moderngl

import moderngl_window


class Cubes(moderngl_window.WindowConfig):
    title = "Cubes"
    resizable = True
    aspect_ratio = 16 / 9
    resource_dir = Path(__file__).parent.resolve() / "resources"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Load the 6 different boxes with different vertex formats
        self.box_v3 = self.load_scene("scenes/box/box-V3F.obj")
        self.box_c3_v3 = self.load_scene("scenes/box/box-C3F_V3F.obj")
        self.box_n3_v3 = self.load_scene("scenes/box/box-N3F_V3F.obj")
        self.box_t2_v3 = self.load_scene("scenes/box/box-T2F_V3F.obj")
        self.box_t2_c3_v3 = self.load_scene("scenes/box/box-T2F_C3F_V3F.obj")
        self.box_t2_n3_v3 = self.load_scene("scenes/box/box-T2F_N3F_V3F.obj")

        self.on_resize(*self.wnd.size)

    def on_render(self, time, frame_time):
        self.ctx.enable_only(moderngl.DEPTH_TEST | moderngl.CULL_FACE)
        rot = glm.mat4(glm.quat(glm.vec3(time, time / 2, time / 3)))

        # Box 1
        view = glm.translate(glm.vec3(-5, 2, -10))
        self.box_v3.draw(self.projection, view * rot)

        # Box 2
        view = glm.translate(glm.vec3(0, 2, -10))
        self.box_c3_v3.draw(self.projection, view * rot)

        # Box 3
        view = glm.translate(glm.vec3(5, 2, -10))
        self.box_n3_v3.draw(self.projection, view * rot)

        # Box 4
        view = glm.translate(glm.vec3(-5, -2, -10))
        self.box_t2_v3.draw(self.projection, view * rot)

        # Box 5
        view = glm.translate(glm.vec3(0, -2, -10))
        self.box_t2_c3_v3.draw(self.projection, view * rot)

        # Box 6
        view = glm.translate(glm.vec3(5, -2, -10))
        self.box_t2_n3_v3.draw(self.projection, view * rot)

    def on_resize(self, width, height):
        super().on_resize(width, height)
        self.projection = glm.perspective(glm.radians(45), self.aspect_ratio, 1, 50)


if __name__ == "__main__":
    Cubes.run()



import random

import moderngl_window
from moderngl_window.utils.scheduler import Scheduler


class CubeSimpleInstancedScheduler(moderngl_window.WindowConfig):
    """Schedule one-off and recurring events"""

    title = "Scheduling example"
    aspect_ratio = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # create a instance of the Scheduler and tell it to use the windows internal timer
        # if we dont pass a timer `time.time()` will be used.
        # by using the internal Timer class the pause and time setting functions
        # will effect our events as well.
        self.scheduler = Scheduler(self.timer)

        # change the color every 1/2 seconds
        color_changing_event = self.scheduler.run_every(self.change_color, 1 / 2)
        # cancel the color changing event after 2 seconds using a priority of 2
        self.scheduler.cancel(color_changing_event, delay=2)
        # restart it after another 2 seconds (4 seconds total)
        color_changing_event = self.scheduler.run_every(self.change_color, 1 / 2, initial_delay=4)

        # after 5 seconds change the window title
        self.scheduler.run_once(self.change_title, 5, arguments=("Changed title",))

    def change_title(self, new_title):
        self.wnd.title = new_title

    def change_color(self):
        self.clear_color = (random.random(), random.random(), random.random(), 0)

    def on_render(self, time: float, frametime: float):
        self.scheduler.execute()


if __name__ == "__main__":
    moderngl_window.run_window_config(CubeSimpleInstancedScheduler)



import moderngl

from moderngl_window.meta import ProgramDescription, ResourceDescription
from moderngl_window.resources.base import BaseRegistry


class Programs(BaseRegistry):
    """Handle program loading"""

    settings_attr = "PROGRAM_LOADERS"
    meta: ProgramDescription

    def resolve_loader(self, meta: ResourceDescription) -> None:
        """Resolve program loader.

        Determines if the references resource is a single
        or multiple glsl files unless ``kind`` is specified.

        Args:
            meta (ProgramDescription): The resource description
        """
        if meta.kind == "":
            if meta.path is None:
                meta.kind = "separate"
            else:
                meta.kind = "single"

        super().resolve_loader(meta)

    def load(self, meta: ResourceDescription) -> moderngl.Program:
        """Loads a shader program with the configured loaders

        Args:
            meta (:py:class:`~moderngl_window.meta.program.ProgramDescription`):
            The resource description
        Returns:
            moderngl.Program: The shader program
        """
        return super().load(meta)


programs = Programs()



import glm
from base import CameraWindow

import moderngl_window
from moderngl_window import geometry


class GeometryBbox(CameraWindow):
    title = "BBox Geometry"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.wnd.mouse_exclusivity = True
        self.prog = self.load_program("scene_default/bbox.glsl")
        self.bbox = geometry.bbox()

        self.prog["color"].value = (1, 1, 1)
        self.prog["bb_min"].value = (-2, -2, -2)
        self.prog["bb_max"].value = (2, 2, 2)
        self.prog["m_model"].write(glm.translate(glm.vec3(0.0, 0.0, -8.0)))

    def on_render(self, time: float, frame_time: float):
        self.ctx.clear()

        self.prog["m_proj"].write(self.camera.projection.matrix)
        self.prog["m_cam"].write(self.camera.matrix)
        self.bbox.render(self.prog)


if __name__ == "__main__":
    moderngl_window.run_window_config(GeometryBbox)



from typing import Any, Generator, Optional, Union

import moderngl_window


class FontMeta:
    """Metdata for texture array"""

    def __init__(self, meta: dict[str, Union[int, list[dict[str, int]]]]):
        self._meta = meta

        assert isinstance(self._meta["characters"], int)
        assert isinstance(self._meta["character_height"], int)
        assert isinstance(self._meta["character_width"], int)
        assert isinstance(self._meta["atlas_height"], int)
        assert isinstance(self._meta["atlas_width"], int)
        assert isinstance(self._meta["character_ranges"], list)

        self.characters = self._meta["characters"]
        self.character_ranges = self._meta["character_ranges"]
        self.character_height = self._meta["character_height"]
        self.character_width = self._meta["character_width"]
        self.atlas_height = self._meta["atlas_height"]
        self.atlas_width = self._meta["atlas_width"]

    @property
    def char_aspect_wh(self) -> float:
        return self.character_width / self.character_height

    def char_aspect_hw(self) -> float:
        return self.character_height / self.character_width


class BaseText:
    """Simple base class for a bitmapped text rendered"""

    def __init__(self) -> None:
        self._meta: Optional[FontMeta] = None
        self._ct: list[int] = []
        self.ctx = moderngl_window.ContextRefs.CONTEXT

    def draw(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError()

    def _translate_string(self, data: str) -> Generator[int, None, None]:
        """Translate string into character texture positions"""
        assert (self._meta is not None) and (
            self._ct is not None
        ), "_meta or _ct (or both) are empty. Did you call _init()?"
        data_bytes = data.encode("iso-8859-1", errors="replace")

        for index, char in enumerate(data_bytes):
            yield self._meta.characters - 1 - self._ct[char]

    def _init(self, meta: FontMeta) -> None:
        self._meta = meta
        # Check if the atlas size is sane
        if not self._meta.characters * self._meta.character_height == self._meta.atlas_height:
            raise ValueError("characters * character_width != atlas_height")

        self._generate_character_map()

    def _generate_character_map(self) -> None:
        """Generate character translation map (latin1 pos to texture pos)"""
        assert self._meta is not None, "You should not call _generate_character_map but _init"
        self._ct = [-1] * 256
        index = 0
        for c_range in self._meta.character_ranges:
            for c_pos in range(c_range["min"], c_range["max"] + 1):
                self._ct[c_pos] = index
                index += 1



from headless import HeadlessTestCase

from moderngl_window import geometry
from moderngl_window.geometry import AttributeNames
from moderngl_window.opengl.vao import BufferInfo


class GeomtryTestCase(HeadlessTestCase):
    custom_attrs = AttributeNames(
        position="test_pos",
        normal="test_normal",
        texcoord_0="test_uv0",
    )

    def test_bbox(self):
        """Create a bounding box"""
        mesh = geometry.bbox(size=(1.0, 1.0, 1.0), name="test_bbox")
        self.assertEqual(mesh.name, "test_bbox")
        self.assertIsInstance(mesh.get_buffer_by_name(AttributeNames.POSITION), BufferInfo)

        # Use custom buffer/attribute names
        mesh = geometry.bbox(size=(1.0, 1.0, 1.0), name="test_bbox", attr_names=self.custom_attrs)
        self.assertIsInstance(mesh.get_buffer_by_name(self.custom_attrs.POSITION), BufferInfo)

    def test_cube(self):
        """Create a cube"""
        mesh = geometry.cube(size=(1.0, 1.0, 1.0), name="test_cube")
        self.assertEqual(mesh.name, "test_cube")
        self.assertIsInstance(mesh.get_buffer_by_name(AttributeNames.POSITION), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(AttributeNames.NORMAL), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(AttributeNames.TEXCOORD_0), BufferInfo)

        # Use custom buffer/attribute names
        mesh = geometry.cube(size=(1.0, 1.0, 1.0), name="test_cube", attr_names=self.custom_attrs)
        self.assertIsInstance(mesh.get_buffer_by_name(self.custom_attrs.POSITION), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(self.custom_attrs.NORMAL), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(self.custom_attrs.TEXCOORD_0), BufferInfo)

    def test_quad_fs(self):
        """Create a fullscreen quad"""
        mesh = geometry.quad_fs(name="test_quad_fs")
        self.assertEqual(mesh.name, "test_quad_fs")
        self.assertIsInstance(mesh.get_buffer_by_name(AttributeNames.POSITION), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(AttributeNames.NORMAL), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(AttributeNames.TEXCOORD_0), BufferInfo)

        # Use custom buffer/attribute names
        mesh = geometry.quad_fs(name="test_quad_fs", attr_names=self.custom_attrs)
        self.assertIsInstance(mesh.get_buffer_by_name(self.custom_attrs.POSITION), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(self.custom_attrs.NORMAL), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(self.custom_attrs.TEXCOORD_0), BufferInfo)

    def test_sphere(self):
        """Create a spwhere"""
        mesh = geometry.sphere(radius=2.0, sectors=32, rings=16, name="test_sphere")
        self.assertEqual(mesh.name, "test_sphere")
        self.assertIsInstance(mesh.get_buffer_by_name(AttributeNames.POSITION), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(AttributeNames.NORMAL), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(AttributeNames.TEXCOORD_0), BufferInfo)

        # Use custom buffer/attribute names
        mesh = geometry.sphere(radius=2.0, sectors=32, rings=16, name="test_sphere", attr_names=self.custom_attrs)
        self.assertIsInstance(mesh.get_buffer_by_name(self.custom_attrs.POSITION), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(self.custom_attrs.NORMAL), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name(self.custom_attrs.TEXCOORD_0), BufferInfo)



import time
from unittest import TestCase

from moderngl_window.timers import clock


class TimerTestCase(TestCase):

    def test_clock_timer(self):
        """Quick and dirty clock timer test"""
        timer = clock.Timer()
        timer.start()
        time.sleep(0.1)
        self.assertFalse(timer.is_paused)
        self.assertTrue(timer.time > 0)
        timer.pause()
        self.assertTrue(timer.is_paused)
        timer.toggle_pause()
        self.assertFalse(timer.is_paused)
        self.assertTrue(timer.is_running)
        timer.time = 10.0
        timer.next_frame()
        pos, duration = timer.stop()
        self.assertTrue(pos >= 10)
        self.assertTrue(duration >= 0)

    def test_not_started(self) -> None:
        """Make sure the timer return 0 when it is never started"""
        timer = clock.Timer()
        t, real_t = timer.stop()
        self.assertTrue(t == 0)
        self.assertTrue(real_t == 0)

    def test_zero_delta(self) -> None:
        """Test that timer handles zero delta gracefully"""
        timer = clock.Timer()
        timer.start()
        # Force a zero delta by setting the same time twice
        timer.time = 1.0
        timer.next_frame()
        timer.time = 1.0
        timer.next_frame()
        # FPS should be 0 when delta is 0 to avoid division by zero
        self.assertEqual(timer.fps, 0)



"""
FIXME: Improve this later. mypy also needs to be included in the linting.
"""

import sys
import shutil
import subprocess
import threading
import webbrowser


def docs():
    """Sphinx documentation"""
    print("Building documentation...")
    subprocess.run("sphinx-build docs build/docs", shell=True)


def clean():
    """Clean up docs and build artifacts"""
    print("Cleaning up...")
    shutil.rmtree("build", ignore_errors=True)


def lint():
    """Run pylint"""
    print("Running pylint...")
    subprocess.run("ruff check", shell=True)


def test():
    """Run tests"""
    print("Running tests...")
    subprocess.run("pytest tests/", shell=True)


def serve():
    """Serve the documentation"""
    print("Serving documentation...")

    def start_server():
        subprocess.run("python -m http.server --directory build/docs", shell=True)

    server_thread = threading.Thread(target=start_server)
    server_thread.start()

    webbrowser.open("http://localhost:8000")
    print("Press Ctrl+C to stop the server")
    server_thread.join()


def run(args: list[str]):
    commands = {
        "html": docs,
        "lint": lint,
        "clean": clean,
        "test": test,
        "serve": serve,
    }
    if len(args) == 0:
        print_help()
        return

    func = commands.get(args[0])
    if func is None:
        print_help()
        return

    func()


def print_help():
    print("Usage: make.py <command>")
    print("Commands:")
    print("  html: Build the documentation")
    print("  lint: Run pylint")
    print("  clean: Clean up docs and build artifacts")
    print("  test: Run tests")
    print("  serve: Serve the documentation")


if __name__ == "__main__":
    run(sys.argv[1:])



from pathlib import Path

import moderngl
from base import OrbitCameraWindow


class OrbitCamCrate(OrbitCameraWindow):
    """
    Example showing how to use a OrbitCamera
    """

    aspect_ratio = 16 / 9
    resource_dir = Path(__file__).parent.resolve() / "resources"
    title = "Crate.obj Model - Orbit Camera"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.wnd.mouse_exclusivity = True

        self.scene = self.load_scene("scenes/crate.obj")

        self.camera.projection.update(near=0.1, far=100.0)
        self.camera.mouse_sensitivity = 0.75
        self.camera.zoom = 2.5

    def on_render(self, time: float, frametime: float):
        self.ctx.enable_only(moderngl.DEPTH_TEST | moderngl.CULL_FACE)

        self.scene.draw(
            projection_matrix=self.camera.projection.matrix,
            camera_matrix=self.camera.matrix,
            time=time,
        )


if __name__ == "__main__":
    OrbitCamCrate.run()



import moderngl
import numpy as np
from PIL import Image

import moderngl_window


class HeadlessTest(moderngl_window.WindowConfig):
    """
    Simple one frame renderer writing to png and exit.
    If you need more fancy stuff, see the custom_config* examples.
    """

    samples = 0  # Headless is not always happy with multisampling

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.wnd.name != "headless":
            raise RuntimeError("This example only works with --window headless option")

        prog = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_vert;
            in vec3 in_color;
            out vec3 color;

            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
                color = in_color;
            }
            """,
            fragment_shader="""
            #version 330

            out vec4 fragColor;
            in vec3 color;

            void main() {
                fragColor = vec4(color, 1.0);
            }
            """,
        )
        # fmt: off
        vertices = np.array([
            -1.0,  -1.0,   1.0, 0.0, 0.0,
            1.0,  -1.0,   0.0, 1.0, 0.0,
            0.0,   1.0,   0.0, 0.0, 1.0],
            dtype='f4',
        )
        # fmt: on
        self.vao = self.ctx.simple_vertex_array(
            prog, self.ctx.buffer(vertices), "in_vert", "in_color"
        )

    def on_render(self, time, frame_time):
        """Render one frame, save to png and close it"""
        # Fill currently bound framebuffer with while background
        self.ctx.clear(1, 1, 1, 1)
        # Render the geometry
        self.vao.render(mode=moderngl.TRIANGLES)

        # Wait for all rendering calls to finish (Might not be needed)
        self.ctx.finish()

        image = Image.frombytes("RGBA", self.wnd.fbo.size, self.wnd.fbo.read(components=4))
        image = image.transpose(Image.FLIP_TOP_BOTTOM)
        image.save("triangle.png", format="png")

        self.wnd.close()


if __name__ == "__main__":
    moderngl_window.run_window_config(HeadlessTest, args=("--window", "headless"))



from pathlib import Path

import glm
import moderngl
from base import CameraWindow

from moderngl_window.scene.camera import KeyboardCamera


class GLTFTest(CameraWindow):
    """
    In oder for this example to work you need to clone the gltf
    model samples repository and ensure resource_dir is set correctly:
    https://github.com/KhronosGroup/glTF-Sample-Models/tree/master/2.0
    """

    title = "GL Transmission Format (glTF) 2.0 Scene"
    window_size = 1280, 720
    aspect_ratio = None
    resource_dir = Path(__file__, "../../../glTF-Sample-Models/2.0").resolve()
    # resource_dir = Path(__file__, "../../tmp/issue_with_draco").resolve()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.wnd.mouse_exclusivity = True

        # self.scene = self.load_scene(
        #   "0147_858530.6583696686_-5537397.529199226_3036197.2162786936.glb", kind="gltf"
        # )

        # --- glTF-Sample-Models ---
        # self.scene = self.load_scene("2CylinderEngine/glTF-Binary/2CylinderEngine.glb")
        # self.scene = self.load_scene("CesiumMilkTruck/glTF-Embedded/CesiumMilkTruck.gltf")
        # self.scene = self.load_scene("CesiumMilkTruck/glTF-Binary/CesiumMilkTruck.glb")
        # self.scene = self.load_scene("CesiumMilkTruck/glTF/CesiumMilkTruck.gltf")
        self.scene = self.load_scene("Sponza/glTF/Sponza.gltf")
        # self.scene = self.load_scene("Lantern/glTF-Binary/Lantern.glb")
        # self.scene = self.load_scene("Buggy/glTF-Binary/Buggy.glb")
        # self.scene = self.load_scene("VC/glTF-Binary/VC.glb")
        # self.scene = self.load_scene("DamagedHelmet/glTF-Binary/DamagedHelmet.glb")
        # self.scene = self.load_scene("BoxInterleaved/glTF/BoxInterleaved.gltf")
        # self.scene = self.load_scene("OrientationTest/glTF/OrientationTest.gltf")
        # self.scene = self.load_scene("AntiqueCamera/glTF/AntiqueCamera.gltf")
        # self.scene = self.load_scene("BoomBox/glTF/BoomBox.gltf")
        # self.scene = self.load_scene("Box/glTF/Box.gltf")
        # self.scene = self.load_scene("BoxTextured/glTF/BoxTextured.gltf")
        # self.scene = self.load_scene(
        #     "BoxTexturedNonPowerOfTwo/glTF/BoxTexturedNonPowerOfTwo.gltf"
        # )
        # self.scene = self.load_scene("BoxVertexColors/glTF/BoxVertexColors.gltf")
        # self.scene = self.load_scene("BrainStem/glTF/BrainStem.gltf")
        # self.scene = self.load_scene("Corset/glTF/Corset.gltf")
        # self.scene = self.load_scene("FlightHelmet/glTF/FlightHelmet.gltf")
        # self.scene = self.load_scene("Fox/glTF/Fox.gltf")
        # self.scene = self.load_scene("GearboxAssy/glTF/GearboxAssy.gltf")
        # self.scene = self.load_scene("ReciprocatingSaw/glTF/ReciprocatingSaw.gltf")
        # self.scene = self.load_scene("RiggedFigure/glTF/RiggedFigure.gltf")
        # self.scene = self.load_scene("RiggedSimple/glTF/RiggedSimple.gltf")
        # self.scene = self.load_scene("SciFiHelmet/glTF/SciFiHelmet.gltf")
        # self.scene = self.load_scene("SimpleMeshes/glTF/SimpleMeshes.gltf")
        # self.scene = self.load_scene("SimpleSparseAccessor/glTF/SimpleSparseAccessor.gltf")
        # self.scene = self.load_scene("Suzanne/glTF/Suzanne.gltf")
        # self.scene = self.load_scene("TextureCoordinateTest/glTF/TextureCoordinateTest.gltf")
        # self.scene = self.load_scene("TextureSettingsTest/glTF/TextureSettingsTest.gltf")
        # self.scene = self.load_scene("VertexColorTest/glTF/VertexColorTest.gltf")
        # self.scene = self.load_scene("WaterBottle/glTF/WaterBottle.gltf")

        # --- Draco compressed ---
        # self.scene = self.load_scene("Box/glTF-Draco/Box.gltf")
        # self.scene = self.load_scene("Buggy/glTF-Draco/Buggy.gltf")
        # self.scene = self.load_scene("2CylinderEngine/glTF-Draco/2CylinderEngine.gltf")
        # self.scene = self.load_scene("CesiumMilkTruck/glTF-Draco/CesiumMilkTruck.gltf")

        self.camera = KeyboardCamera(
            self.wnd.keys,
            fov=75.0,
            aspect_ratio=self.wnd.aspect_ratio,
            near=0.1,
            far=1000.0,
        )
        self.camera.velocity = 10.0
        self.camera.mouse_sensitivity = 0.25

        # Use this for gltf scenes for better camera controls
        # if self.scene.diagonal_size > 0:
        #     self.camera.velocity = self.scene.diagonal_size / 5.0

        self.camera.position = (
            self.scene.get_center()
            + glm.vec3(0.0, 0.0, self.scene.diagonal_size / 1.75)
        )

    def on_render(self, time: float, frame_time: float):
        """Render the scene"""
        self.ctx.enable_only(moderngl.DEPTH_TEST | moderngl.CULL_FACE)

        self.scene.draw(
            projection_matrix=self.camera.projection.matrix,
            camera_matrix=self.camera.matrix,
            time=time,
        )

        # # Draw bounding boxes
        # self.scene.draw_bbox(
        #     projection_matrix=self.camera.projection.matrix,
        #     camera_matrix=self.camera.matrix,
        #     children=True,
        #     color=(0.75, 0.75, 0.75),
        # )

        # self.scene.draw_wireframe(
        #     projection_matrix=self.camera.projection.matrix,
        #     camera_matrix=self.camera.matrix,
        #     color=(1, 1, 1, 1),
        # )


if __name__ == "__main__":
    GLTFTest.run()



"""
GPU Version of https://github.com/salt-die/ripple

Hold left mouse button to place drop in the surface
"""

import random
from pathlib import Path

import moderngl
import numpy as np

import moderngl_window
from moderngl_window import geometry, screenshot


class Water(moderngl_window.WindowConfig):
    title = "Water"
    resource_dir = (Path(__file__) / "../../resources").absolute()
    aspect_ratio = None  # We'll do manual viewport for now
    window_size = 1280, 720
    resizable = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.size = self.wnd.buffer_size
        self.viewport = (0, 0, self.size[0], self.size[1])

        self.quad_fs = geometry.quad_fs()
        self.sprite = geometry.quad_2d(size=(9 / self.wnd.size[0], 9 / self.wnd.size[1]))

        self.texture_1 = self.ctx.texture(self.size, components=3)
        self.texture_2 = self.ctx.texture(self.size, components=3)

        self.fbo_1 = self.ctx.framebuffer(color_attachments=[self.texture_1])
        self.fbo_1.viewport = self.viewport
        self.fbo_2 = self.ctx.framebuffer(color_attachments=[self.texture_2])
        self.fbo_2.viewport = self.viewport

        # fmt: off
        drop = np.array([
            [0.0, 0.0, 1/6, 1/5, 1/4, 1/5, 1/6, 0.0, 0.0],
            [0.0, 1/6, 1/5, 1/4, 1/3, 1/4, 1/5, 1/6, 0.0],
            [1/6, 1/5, 1/4, 1/3, 1/2, 1/3, 1/4, 1/5, 1/6],
            [1/5, 1/4, 1/3, 1/2, 1.0, 1/2, 1/3, 1/4, 1/5],
            [1/4, 1/3, 1/2, 1.0, 1.0, 1.0, 1/2, 1/3, 1/4],
            [1/5, 1/4, 1/3, 1/2, 1.0, 1/2, 1/3, 1/4, 1/5],
            [1/6, 1/5, 1/4, 1/3, 1/2, 1/3, 1/4, 1/5, 1/6],
            [0.0, 1/6, 1/5, 1/4, 1/3, 1/4, 1/5, 1/6, 0.0],
            [0.0, 0.0, 1/6, 1/5, 1/4, 1/5, 1/6, 0.0, 0.0],
        ])
        # fmt: on

        self.drops_texture = self.ctx.texture((9, 9), components=1, dtype="f4")
        self.drops_texture.write(drop.astype("f4").tobytes())

        # programs
        self.drop_program = self.load_program("programs/water/drop.glsl")
        self.wave_program = self.load_program("programs/water/wave.glsl")
        self.texture_program = self.load_program("programs/water/texture.glsl")
        self.wave_program["texture0"].value = 0
        self.wave_program["texture1"].value = 1

        self.mouse_pos = 0, 0
        self.wnd.fbo.viewport = self.viewport

    def on_render(self, time, frame_time):
        # randomize color
        self.drop_program["color"].value = random.random(), random.random(), random.random()

        self.fbo_2.use()

        # Render drop (with additive blending) when mouse is pressed
        if self.wnd.mouse_states.any:
            self.ctx.enable(moderngl.BLEND)
            self.ctx.blend_func = moderngl.ONE, moderngl.ONE
            self.drops_texture.use()
            self.drop_program["pos"].value = self.mouse_pos
            self.sprite.render(self.drop_program)
            self.ctx.disable(moderngl.BLEND)

        # HACK: Just draw 100 new drops per frame (copy paste from above)
        # This is pretty terrible and slow!
        self.ctx.enable(moderngl.BLEND)
        self.ctx.blend_func = moderngl.ONE, moderngl.ONE
        self.drops_texture.use()
        for i in range(10):
            self.drop_program["pos"].value = random.random() * 2 - 1.0, random.random() * 2 - 1
            self.sprite.render(self.drop_program)
        self.ctx.disable(moderngl.BLEND)

        self.fbo_1.use()

        # Process the water
        self.texture_2.use(location=0)
        self.texture_1.use(location=1)
        self.quad_fs.render(self.wave_program)

        # Render the result to the screen.
        # We can blit only when the texture format matches the default framebuffer
        self.wnd.fbo.use()
        self.texture_1.use()
        self.quad_fs.render(self.texture_program)

        # Swap texture and fbo
        self.texture_1, self.texture_2 = self.texture_2, self.texture_1
        self.fbo_1, self.fbo_2 = self.fbo_2, self.fbo_1

    def on_mouse_position_event(self, x, y, dx, dy):
        xx = x * 2 / self.wnd.size[0] - 1.0
        yy = -y * 2 / self.wnd.size[1] + 1.0
        self.mouse_pos = xx, yy

    def on_mouse_drag_event(self, x, y, dx, dy):
        self.mouse_position_event(x, y, dx, dy)

    def on_key_event(self, key, action, modifiers):
        keys = self.wnd.keys

        # Key presses
        if action == keys.ACTION_PRESS:
            if key == keys.F1:
                screenshot.create(self.fbo_1)


if __name__ == "__main__":
    moderngl_window.run_window_config(Water)



import logging
from pathlib import Path
from typing import Optional, Union

import moderngl

from moderngl_window.exceptions import ImproperlyConfigured
from moderngl_window.loaders.base import BaseLoader
from moderngl_window.opengl import program

logger = logging.getLogger(__name__)


class Loader(BaseLoader):
    kind = "separate"
    meta: program.ProgramDescription

    def load(
        self,
    ) -> Union[moderngl.Program, moderngl.ComputeShader, program.ReloadableProgram]:
        """Loads a shader program were each shader is a separate file.

        This detected and dictated by the ``kind`` in the ``ProgramDescription``.

        Returns:
            moderngl.Program: The Program instance
        """
        prog: Union[moderngl.Program, moderngl.ComputeShader, program.ReloadableProgram]

        vs_source = self._load_shader("vertex", self.meta.vertex_shader)
        geo_source = self._load_shader("geometry", self.meta.geometry_shader)
        fs_source = self._load_shader("fragment", self.meta.fragment_shader)
        tc_source = self._load_shader("tess_control", self.meta.tess_control_shader)
        te_source = self._load_shader("tess_evaluation", self.meta.tess_evaluation_shader)
        cs_source = self._load_shader("compute", self.meta.compute_shader)

        if vs_source:
            shaders = program.ProgramShaders.from_separate(
                self.meta,
                vs_source,
                geometry_source=geo_source,
                fragment_source=fs_source,
                tess_control_source=tc_source,
                tess_evaluation_source=te_source,
            )
            shaders.handle_includes(self._load_source)
            prog = shaders.create()

            # Wrap the program if reloadable is set
            if self.meta.reloadable:
                # Disable reload flag so reloads will return Program instances
                self.meta.reloadable = False
                # Wrap it ..
                prog = program.ReloadableProgram(self.meta, prog)
        elif cs_source:
            shaders = program.ProgramShaders.compute_shader(self.meta, cs_source)
            shaders.handle_includes(self._load_source)
            prog = shaders.create_compute_shader()
        else:
            raise ImproperlyConfigured("Cannot find a shader source to load")

        return prog

    def _load_shader(self, shader_type: str, path: Optional[str]) -> Optional[str]:
        """Load a single shader source"""
        if path is not None:
            resolved_path = self.find_program(path)
            if not resolved_path:
                raise ImproperlyConfigured("Cannot find {} shader '{}'".format(shader_type, path))

            logger.info("Loading: %s", resolved_path)

            with open(str(resolved_path), "r") as fd:
                return fd.read()
        return None

    def _load_source(self, path: Union[Path, str]) -> tuple[Path, str]:
        """Finds and loads a single source file.

        Args:
            path: Path to resource
        Returns:
            tuple[resolved_path, source]: The resolved path and the source
        """
        resolved_path = self.find_program(path)
        if resolved_path is None:
            raise ImproperlyConfigured("Cannot find program '{}'".format(path))

        logger.info("Loading: %s", path)

        with open(str(resolved_path), "r") as fd:
            return resolved_path, fd.read()



from pathlib import Path
from unittest import TestCase

from moderngl_window import resources
from moderngl_window.meta import DataDescription
from moderngl_window.opengl import program

resources.register_dir((Path(__file__).parent / 'fixtures/resources').resolve())


class ShaderSourceTestCase(TestCase):
    maxDiff = None

    def test_no_version(self):
        """Missing #version statement should raise a shader error"""
        with self.assertRaises(program.ShaderError):
            program.ShaderSource(
                program.VERTEX_SHADER,
                "test.glsl",
                """
                in vec2 in_pos;
                out vec2 out_pos;
                void main() {
                    out_pos = in_pos;
                }
                """
            )

    def test_no_version_first_line(self):
        """A shader should always start have a #version statement on the first line"""
        with self.assertRaises(program.ShaderError):
            program.ShaderSource(
                program.VERTEX_SHADER,
                "test.glsl",
                """
                in vec2 in_pos;
                out vec2 out_pos;
                # version 330
                void main() {
                    out_pos = in_pos;
                }
                """
            )

    def test_version_strip(self):
        """Make sure we can start with blank lines before #version"""
        program.ShaderSource(
            program.VERTEX_SHADER,
            "test.glsl",
            """

            #version 330
            in vec2 in_pos;
            out vec2 out_pos;
            void main() {
                out_pos = in_pos;
            }
            """
        )

    def test_define(self):
        """inject a define value"""
        shader = program.ShaderSource(
            program.VERTEX_SHADER,
            "test.glsl",
            """
            #version 330
            in vec2 in_pos;
            out vec2 out_pos;
            #define NUM_THINGS 16
            #define SCALE = 1.0
            void main() {
                out_pos = in_pos * SCALE;
            }
            """,
            defines={'NUM_THINGS': '100', 'SCALE': '2.0'}
        )
        self.assertTrue("#define NUM_THINGS 100" in shader.source)
        self.assertTrue("#define SCALE 2.0" in shader.source)

    def test_define_compute(self):
        """Injecting defines in compute shader"""
        shader = program.ShaderSource(
            program.COMPUTE_SHADER,
            "compute.glsl",
            """
            #version 430
            #define TEST 0
            #define THING 0
            layout (local_size_x = 1, local_size_y = 1) in;
            layout (std430, binding = 1) buffer Input {
                float v1[4];
            };
            layout (std430, binding = 2) buffer Output {
                float v2[4];
            };
            uniform float mul;
            uniform vec4 add;
            void main() {
                v2[0] = v1[3] * mul + add.x;
                v2[1] = v1[2] * mul + add.y;
                v2[2] = v1[1] * mul + add.z;
                v2[3] = v1[0] * mul + add.w;
            }
            """,
            defines={'TEST': 12, 'THING': '23'},
        )
        self.assertTrue("#define TEST 12" in shader.source)
        self.assertTrue("#define THING 23" in shader.source)

    def test_out_attributes(self):
        """Ensure out attributes are proprely parsed"""
        shader = program.ShaderSource(
            program.VERTEX_SHADER,
            "test.glsl",
            """
            #version 330
            out vec2 out_pos;
            out vec2 out_vel;
            void main() {
                out_pos = vec2(1.0);
                out_vel = vec2(1.0);
            }
            """,
        )
        assert shader.find_out_attribs() == ['out_pos', 'out_vel']
        # With layout qualifiers
        shader = program.ShaderSource(
            program.VERTEX_SHADER,
            "test.glsl",
            """
            #version 330
            layout(location = 0) out vec2 out_pos;
            layout(location = 1) out vec2 out_vel;
            void main() {
                out_pos = vec2(1.0);
                out_vel = vec2(1.0);
            }
            """,
        )
        assert shader.find_out_attribs() == ['out_pos', 'out_vel']

    def test_include(self):
        """Test #include preprocessors (recursive)"""
        def load_source(path):
            return path, resources.data.load(DataDescription(path, kind='text'))

        path = 'programs/include_test.glsl'
        source = load_source(path)[1]
        source_vs = program.ShaderSource(program.VERTEX_SHADER, path, source, defines={'TEST': '100'})
        source_vs.handle_includes(load_source)

        self.assertEqual(source_vs.source, INCLUDE_RESULT)
        self.assertEqual(source_vs.source_list[0].name, 'programs/include_test.glsl')
        self.assertEqual(source_vs.source_list[0].id, 0)
        self.assertEqual(source_vs.source_list[1].name, 'programs/includes/blend_functions.glsl')
        self.assertEqual(source_vs.source_list[1].id, 1)
        self.assertEqual(source_vs.source_list[2].name, 'programs/includes/utils.glsl')
        self.assertEqual(source_vs.source_list[2].id, 2)
        self.assertEqual(source_vs.source_list[3].name, 'programs/includes/utils_1.glsl')
        self.assertEqual(source_vs.source_list[3].id, 3)
        self.assertEqual(source_vs.source_list[4].name, 'programs/includes/utils_2.glsl')
        self.assertEqual(source_vs.source_list[4].id, 4)

    def test_include_circular(self):
        """Ensure circular includes are detected"""
        def load_source(path):
            return path, resources.data.load(DataDescription(path, kind='text'))

        path = 'programs/include_test_circular.glsl'
        _, source = load_source(path)
        source = program.ShaderSource(program.VERTEX_SHADER, path, source)
        with self.assertRaises(program.ShaderError):
            source.handle_includes(load_source)


INCLUDE_RESULT = """#version 330
#define VERTEX_SHADER 1
#line 2

vec3 blendMultiply(vec3 base, vec3 blend) {
	return base * blend;
}

vec3 blendMultiply(vec3 base, vec3 blend, float opacity) {
	return (blendMultiply(base, blend) * opacity + base * (1.0 - opacity));
}

// Utils 1
#define TEST 100

// Utils 2
#define TEST 100



#define TEST 100

#if defined VERTEX_SHADER

in vec3 in_position;

void main() {
    gl_Position = vec4(in_position, 1.0);
}


#elif defined FRAGMENT_SHADER

out vec4 fragColor;

void main() {
    fragColor = vec4(1.0);
}

#endif"""



"""
GPU version of https://github.com/salt-die/navier-stokes

Interact with the window using to add pressure and momentum.
* Left mouse button (including drag) pokes the surface
* Right mouse button places walls
"""

import random
from pathlib import Path

import glm

import moderngl_window
from moderngl_window import geometry

external_flow = 0.35
viscosity = 0.018  # Is it odd that negative viscosity still works?
rho = 1.06  # Density
damping = 0.999


class NavierStokes2D(moderngl_window.WindowConfig):
    title = "Navier Stokes 2D"
    resource_dir = (Path(__file__) / "../../resources").absolute()
    # window_size = 3440, 1440
    # window_size = 512, 512
    aspect_ratio = None  # Aspect ratio should change with window size
    resizable = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # various vars
        self.m_proj = glm.ortho(
            0,
            self.wnd.buffer_width,
            0,
            self.wnd.buffer_height,
            -1,
            1,
        )
        size = self.wnd.buffer_size

        # Geometry
        self.quad_fs = geometry.quad_fs()
        sprite_size = 9 * 4
        self.drop_geometry = geometry.quad_2d(size=(sprite_size, sprite_size))
        self.wall_geometry = geometry.quad_2d(size=(sprite_size, sprite_size))

        # Framebuffers
        self.momentum_texture_1 = self.ctx.texture(size, 1, dtype="f4")
        self.momentum_texture_1.repeat_x = True
        self.momentum_texture_1.repeat_y = True
        self.momentum_texture_2 = self.ctx.texture(size, 1, dtype="f4")
        self.momentum_texture_2.repeat_x = True
        self.momentum_texture_2.repeat_y = True
        self.pressure_texture_1 = self.ctx.texture(size, 1, dtype="f4")
        self.pressure_texture_2 = self.ctx.texture(size, 1, dtype="f4")
        self.walls_texture = self.ctx.texture(size, 1, dtype="f4")
        self.difference_texture = self.ctx.texture(size, 1, dtype="f4")

        self.momentum_fbo_1 = self.ctx.framebuffer(color_attachments=[self.momentum_texture_1])
        self.momentum_fbo_2 = self.ctx.framebuffer(color_attachments=[self.momentum_texture_2])
        self.pressure_fbo_1 = self.ctx.framebuffer(color_attachments=[self.pressure_texture_1])
        self.pressure_fbo_2 = self.ctx.framebuffer(color_attachments=[self.pressure_texture_2])
        self.walls_fbo = self.ctx.framebuffer(color_attachments=[self.walls_texture])
        self.difference_fbo = self.ctx.framebuffer(color_attachments=[self.difference_texture])

        # programs
        self.texture_prog = self.load_program("programs/navier-stokes/texture.glsl")
        self.drop_prog = self.load_program("programs/navier-stokes/drop.glsl")
        self.drop_prog["m_proj"].write(self.m_proj)
        self.combine_prog = self.load_program("programs/navier-stokes/combine.glsl")
        self.combine_prog["pressure_texture"].value = 0
        self.combine_prog["wall_texture"].value = 1
        self.momentum_prog = self.load_program("programs/navier-stokes/momentum.glsl")
        self.momentum_prog["rho"].value = rho
        self.momentum_prog["damping"].value = damping
        self.momentum_prog["external_flow"].value = external_flow
        self.momentum_prog["momentum_texture"].value = 0
        self.momentum_prog["pressure_texture"].value = 1
        self.momentum_prog["walls_texture"].value = 3
        self.flow_prog = self.load_program("programs/navier-stokes/flow.glsl")
        self.flow_prog["external_flow"].value = external_flow
        self.poisson_prog = self.load_program("programs/navier-stokes/poisson.glsl")
        self.pressure_prog = self.load_program("programs/navier-stokes/pressure.glsl")
        self.pressure_prog["rho"].value = rho
        self.pressure_prog["damping"].value = damping
        self.pressure_prog["pressure_texture"].value = 0
        self.pressure_prog["difference_texture"].value = 1
        self.pressure_prog["walls_texture"].value = 3

        self.reset()

    def reset(self):
        self.momentum_fbo_2.clear()
        self.pressure_fbo_2.clear()
        self.walls_fbo.clear()

    def drop(self, x, y):
        rng = 40
        for i in range(10):
            pos = (
                x + int((random.random() - 0.5) * rng),
                self.wnd.buffer_size[1] - y + int((random.random() - 0.5) * rng),
            )
            # Render drop into pressure texture
            self.pressure_fbo_2.use()
            self.drop_prog["pos"].value = pos
            self.drop_prog["force"].value = random.random() * 2.0
            self.drop_prog["write_value"].value = 1.0
            # self.drop.use()
            self.drop_geometry.render(self.drop_prog)
            self.momentum_fbo_2.use()
            self.drop_prog["write_value"].value = 0.0
            self.drop_geometry.render(self.drop_prog)

    def wall(self, x, y):
        pos = x, self.wnd.buffer_height - y
        self.walls_fbo.use()
        self.drop_prog["pos"].value = pos
        self.drop_prog["force"].value = 1.0
        self.drop_prog["write_value"].value = 1.0
        self.drop_geometry.render(self.drop_prog)

    def on_render(self, time: float, frame_time: float):
        # calculate momentum
        self.momentum_fbo_1.use()
        self.momentum_texture_2.use(location=0)
        self.pressure_texture_2.use(location=1)
        self.walls_texture.use(location=3)
        self.quad_fs.render(self.momentum_prog)

        # External Flow
        self.momentum_swap()
        self.momentum_fbo_1.use()
        self.momentum_texture_2.use()
        self.quad_fs.render(self.flow_prog)

        # change in momentum
        # .. poisson momentum
        self.difference_fbo.use()
        self.momentum_texture_1.use()
        self.quad_fs.render(self.poisson_prog)
        # .. pressure
        self.pressure_fbo_1.use()
        self.pressure_texture_2.use(location=0)
        self.difference_texture.use(location=1)
        self.walls_texture.use(location=3)
        self.quad_fs.render(self.pressure_prog)

        # Render final result
        self.wnd.fbo.use()
        self.pressure_texture_1.use(location=0)
        self.walls_texture.use(location=1)
        self.quad_fs.render(self.combine_prog)

        self.momentum_swap()
        self.pressure_swap()

    def momentum_swap(self):
        self.momentum_texture_1, self.momentum_texture_2 = (
            self.momentum_texture_2,
            self.momentum_texture_1,
        )
        self.momentum_fbo_1, self.momentum_fbo_2 = (
            self.momentum_fbo_2,
            self.momentum_fbo_1,
        )

    def pressure_swap(self):
        self.pressure_texture_1, self.pressure_texture_2 = (
            self.pressure_texture_2,
            self.pressure_texture_1,
        )
        self.pressure_fbo_1, self.pressure_fbo_2 = (
            self.pressure_fbo_2,
            self.pressure_fbo_1,
        )

    def on_mouse_press_event(self, x, y, button):
        if button == self.wnd.mouse.left:
            self.drop(x, y)
        elif button == self.wnd.mouse.right:
            self.wall(x, y)

    def on_mouse_drag_event(self, x, y, dx, dy):
        if self.wnd.mouse_states.left:
            self.drop(x, y)
        elif self.wnd.mouse_states.right:
            self.wall(x, y)

    def on_key_event(self, key, action, modifiers):
        keys = self.wnd.keys

        if action == keys.ACTION_PRESS:
            if key == keys.R:
                self.reset()


if __name__ == "__main__":
    NavierStokes2D.run()



from moderngl_window.context.base.keys import BaseKeys as BaseKeys
from moderngl_window.context.base.keys import KeyModifiers as KeyModifiers
from moderngl_window.context.base.window import BaseWindow as BaseWindow
from moderngl_window.context.base.window import WindowConfig as WindowConfig

__all__ = [
    "BaseKeys",
    "KeyModifiers",
    "BaseWindow",
    "WindowConfig",
]



"""
Based on BlubberQuark's blog:
https://blubberquark.tumblr.com/post/185013752945/using-moderngl-for-post-processing-shaders-with

Clears the screen using opengl with a constantly changing
color value and alpha blend a pygame surface on top.
"""

import math
from pathlib import Path

import moderngl
import pygame

import moderngl_window
from moderngl_window import geometry


class Pygame(moderngl_window.WindowConfig):
    """
    Example using pygame with moderngl.
    Needs to run with ``--window pygame2`` option.
    """

    title = "Pygame"
    window_size = 1280, 720
    resource_dir = (Path(__file__) / "../../resources").absolute()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Create a 24bit (rgba) offscreen surface pygame can render to
        self.pg_screen = pygame.Surface(self.window_size, flags=pygame.SRCALPHA)
        # 24 bit (rgba) moderngl texture
        self.pg_texture = self.ctx.texture(self.window_size, 4)
        self.pg_texture.filter = moderngl.NEAREST, moderngl.NEAREST
        self.pg_texture.swizzle = "BGRA"

        self.texture_program = self.load_program("programs/texture.glsl")
        self.quad_texture = self.load_texture_2d("textures/python-bg.png")
        self.quad_fs = geometry.quad_fs()

    def on_render(self, time: float, frametime: float):
        self.ctx.clear()

        self.ctx.enable(moderngl.BLEND)

        # Render background graphics
        self.quad_texture.use()
        self.texture_program["texture0"].value = 0
        self.quad_fs.render(self.texture_program)

        # Render foreground objects
        self.pg_texture.use()
        self.render_pygame(time)
        self.quad_fs.render(self.texture_program)

        self.ctx.disable(moderngl.BLEND)

    def render_pygame(self, time):
        """Render to offscreen surface and copy result into moderngl texture"""
        self.pg_screen.fill((0, 0, 0, 0))  # Make sure we clear with alpha 0!
        N = 8
        for i in range(N):
            time_offset = 6.28 / N * i
            pygame.draw.circle(
                self.pg_screen,
                ((i * 50) % 255, (i * 100) % 255, (i * 20) % 255),
                (
                    math.sin(time + time_offset) * 200 + self.window_size[0] // 2,
                    math.cos(time + time_offset) * 200 + self.window_size[1] // 2,
                ),
                math.sin(time) * 7 + 15,
            )

        # Get the buffer view of the Surface's pixels
        # and write this data into the texture
        texture_data = self.pg_screen.get_view("1")
        self.pg_texture.write(texture_data)


if __name__ == "__main__":
    moderngl_window.run_window_config(Pygame, args=("--window", "pygame2"))



"""
Video player implementation using PyAV for decoding and ModernGL for rendering.
Requires: pip install av
"""

import logging
from pathlib import Path
from typing import Union, Literal

import av
import moderngl
import moderngl_window
from moderngl_window import geometry

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class VideoDecoder:
    """Handles video decoding using PyAV."""

    def __init__(self, path: Union[str, Path]):
        self._path = Path(path)
        if not self._path.exists():
            raise FileNotFoundError(f"Video file not found: {self._path}")

        self.container = av.open(str(self._path))
        self.video = self.container.streams.video[0]
        self.video.thread_type = "AUTO"
        self.video.codec_context.pix_fmt = "yuv420p"

        self._current_frame = 0

    @property
    def duration(self) -> float:
        """Video duration in seconds."""
        if self.video.duration is None:
            return float(self.container.duration * float(self.container.time_base))
        return float(self.video.duration * float(self.video.time_base))

    @property
    def framerate(self) -> float:
        """Average framerate."""
        rate = self.video.average_rate
        return rate.numerator / rate.denominator

    @property
    def size(self) -> tuple[int, int]:
        """Video dimensions (width, height)."""
        return self.video.width, self.video.height

    def seek(self, time_seconds: float) -> None:
        """Seek to specified time position."""
        try:
            time_seconds = max(0, min(time_seconds, self.duration))
            timestamp = int(time_seconds / float(self.video.time_base))
            self.container.seek(timestamp, stream=self.video)
            self._current_frame = int(time_seconds * self.framerate)
            self.container.decode(video=0)
        except Exception as e:
            logger.error(f"Seek failed: {e}")
            self.container.seek(0)
            self._current_frame = 0

    def get_frames(self):
        """Generate video frames from current position."""
        try:
            for packet in self.container.demux(video=0):
                for frame in packet.decode():
                    self._current_frame += 1
                    yield frame.to_rgb().planes[0]
        except Exception as e:
            logger.error(f"Frame generation error: {e}")

    @property
    def frames(self) -> int:
        """Total number of frames in video."""
        return int(self.duration * self.framerate)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.container.close()

    def close(self):
        """Explicitly close the video container."""
        self.container.close()


class VideoPlayer:
    """Handles video playback and rendering using ModernGL."""

    # Add constants at class level
    FRAME_DIFF_THRESHOLD = 5
    MAX_BEHIND_COUNT = 10
    SKIP_OFFSET = 2

    def __init__(self, ctx: moderngl.Context, path: Union[str, Path]):
        self._ctx = ctx
        self._decoder = VideoDecoder(path)
        self._texture = self._ctx.texture(self._decoder.size, 3, dtype="f1")
        self._frames = self._decoder.get_frames()

        self._current_frame = 0
        self._target_frame = 0
        self._behind_count = 0
        self._paused = False

    @property
    def fps(self) -> float:
        return self._decoder.framerate

    @property
    def duration(self) -> float:
        return self._decoder.duration

    @property
    def frames(self) -> int:
        return self._decoder.frames

    @property
    def size(self) -> tuple[int, int]:
        return self._decoder.size

    @property
    def texture(self) -> moderngl.Texture:
        return self._texture

    def update(self, time: float) -> bool:
        """Update video playback state."""
        if self._paused:
            return False

        self._target_frame = int(time * self.fps)
        frame_diff = self._target_frame - self._current_frame

        # Check if we've reached the end
        if self._current_frame >= self.frames:
            self.seek(0)
            return True

        # Handle falling behind
        if frame_diff > self.FRAME_DIFF_THRESHOLD:
            self._behind_count += 1
            skip_to = min(self._target_frame - self.SKIP_OFFSET, self.frames - 1)

            if self._behind_count > self.MAX_BEHIND_COUNT:
                self._decoder.seek(skip_to / self.fps)
                self._frames = self._decoder.get_frames()
                self._current_frame = skip_to
                self._behind_count = 0
            else:
                try:
                    while self._current_frame < skip_to:
                        next(self._frames)
                        self._current_frame += 1
                except StopIteration:
                    self.seek(0)
                    return True
        else:
            self._behind_count = max(0, self._behind_count - 1)
            if frame_diff > 0:
                try:
                    data = next(self._frames)
                    self._texture.write(data)
                    self._current_frame += 1
                except StopIteration:
                    self.seek(0)
                    return True

        return False

    def seek(self, time: float) -> None:
        """Seek to specified time position."""
        time = max(0, min(time, self.duration))
        self._decoder.seek(time)
        self._frames = self._decoder.get_frames()
        self._current_frame = int(time * self.fps)
        self._target_frame = self._current_frame
        self._behind_count = 0

    def toggle_pause(self) -> None:
        """Toggle pause state."""
        self._paused = not self._paused

    @property
    def current_frame(self) -> int:
        return self._current_frame

    @property
    def target_frame(self) -> int:
        return self._target_frame

    @property
    def is_paused(self) -> bool:
        return self._paused


class VideoPlayerWindow(moderngl_window.WindowConfig):
    """ModernGL window configuration for video playback."""

    gl_version = (3, 3)
    title = "Video Player"
    resource_dir = Path(__file__).parent.resolve() / "resources"
    vsync = True
    seek_time = 1.0  # Seconds to seek when using arrow keys
    log_level = logging.DEBUG

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Initialize video player
        video_path = self.resource_dir / "videos" / "Lightning - 33049.mp4"

        self.player = VideoPlayer(self.ctx, video_path)

        # Setup rendering
        self.quad = geometry.quad_fs()
        self.program = self.load_program("programs/texture_flipped.glsl")

        # Setup stats printing
        self._last_print_time = 0

    def on_render(self, time: float, frametime: float) -> None:
        """Render frame."""
        if self.player.update(time):  # Check if video ended
            self.timer.time = 0  # Reset timer if video ended

        # Render video
        self.player.texture.use(0)
        self.quad.render(self.program)

        # Print debug stats every 0.5 seconds regardless of pause state
        if (time - self._last_print_time) >= 0.5 or time < self._last_print_time:
            # Get FPS values with safety checks
            fps_avg = self.timer.fps_average if self.timer.time > 0 else 0.0

            logger.debug(
                "Movie Target FPS: %.1f | Window FPS: %.1f | Frame: %d/%d | \
                Time: %.2f/%.2f | Frame Diff: %d | Paused: %s",
                self.player.fps,
                fps_avg,
                self.player.current_frame,
                self.player.frames,
                self.timer.time,
                self.player.duration,
                self.player.target_frame - self.player.current_frame,
                self.player.is_paused,
            )
            self._last_print_time = time

    def _handle_seek(self, direction: Literal["forward", "backward"]) -> None:
        """Handle seeking in video. direction: 'forward' or 'backward'"""
        seek_amount = self.seek_time if direction == "forward" else -self.seek_time
        new_time = max(0, min(self.player.duration, self.timer.time + seek_amount))

        if self.timer.is_paused:
            self.player.seek(new_time)
        else:
            self.timer.time = new_time
            self.player.seek(new_time)

    def on_key_event(self, key, action, modifiers) -> None:
        """Handle keyboard input."""
        super().on_key_event(key, action, modifiers)
        keys = self.wnd.keys

        if action == keys.ACTION_PRESS:
            if key == keys.LEFT:
                self._handle_seek("backward")
            elif key == keys.RIGHT:
                self._handle_seek("forward")
            elif key == keys.SPACE:
                self.timer.toggle_pause()
                self.player.toggle_pause()


if __name__ == "__main__":
    VideoPlayerWindow.run()



import gzip
from pathlib import Path

import moderngl
import numpy
import trimesh

from moderngl_window.exceptions import ImproperlyConfigured
from moderngl_window.loaders.base import BaseLoader
from moderngl_window.opengl.vao import VAO
from moderngl_window.scene import Material, Mesh, Node, Scene


class Loader(BaseLoader):
    kind = "stl"
    file_extensions = [
        [".stl"],
        [".stl", ".gz"],
    ]

    def load(self) -> Scene:
        """Loads and stl scene/file

        Returns:
            Scene: The Scene instance
        """
        path = self.find_scene(self.meta.path)
        if not path:
            raise ImproperlyConfigured("Scene '{}' not found".format(self.meta.path))

        file_obj = str(path)
        if file_obj.endswith(".gz"):
            file_obj = gzip.GzipFile(file_obj)

        stl_mesh = trimesh.load(file_obj, file_type="stl")
        path = self.meta.resolved_path
        if isinstance(path, Path):
            resolved = path.as_posix()
        else:
            resolved = None
        scene = Scene(resolved)
        scene_mesh = Mesh("mesh")
        scene_mesh.material = Material("default")

        vao = VAO("mesh", mode=moderngl.TRIANGLES)
        vao.buffer(numpy.array(stl_mesh.vertices, dtype="f4"), "3f", ["in_position"])
        vao.buffer(numpy.array(stl_mesh.vertex_normals, dtype="f4"), "3f", ["in_normal"])
        vao.index_buffer(numpy.array(stl_mesh.faces, dtype="u4"))
        scene_mesh.vao = vao
        scene_mesh.add_attribute("POSITION", "in_position", 3)
        scene_mesh.add_attribute("NORMAL", "in_normal", 3)

        scene.meshes.append(scene_mesh)
        scene.root_nodes.append(Node(mesh=scene_mesh))
        scene.prepare()

        return scene



import time
from unittest import TestCase

from moderngl_window.timers.clock import Timer
from moderngl_window.utils.scheduler import Scheduler


class SchedulingTestCase(TestCase):
    def set_value(self, v):
        self.test_value = v

    def increase_value(self):
        self.test_value += 1

    def test_clock_timer(self):
        """Quick and dirty scheduling test"""
        timer = Timer()
        timer.start()

        scheduler = Scheduler(timer)
        self.test_value = False
        scheduler.run_once(self.set_value, 0.1, arguments=(True,))
        time.sleep(0.11)
        scheduler.execute()
        self.assertTrue(self.test_value)

        self.test_value = 0
        event = scheduler.run_every(self.increase_value, 0.1)
        start = time.time()
        delta = time.time()
        while (delta - start) <= 0.5:
            # simulate a render loop
            print(self.test_value)
            time.sleep(0.01)
            scheduler.execute()
            delta = time.time()
        self.assertEqual(self.test_value, 5)
        # make sure it stays 5 once we cancel the event
        scheduler.cancel(event)
        time.sleep(0.11)
        scheduler.execute()
        self.assertEqual(self.test_value, 5)



from pathlib import Path

import moderngl
from base import CameraWindow

from moderngl_window import geometry


class Cubemap(CameraWindow):
    """Example loading and rendering a cubemap"""

    title = "Skybox with cubemap"
    resource_dir = (Path(__file__).parent / "resources").resolve()
    aspect_ratio = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.wnd.mouse_exclusivity = True
        self.cube = geometry.cube(size=(20, 20, 20))
        self.texture = self.load_texture_cube(
            neg_x="textures/cubemaps/yokohama/posx.jpg",
            neg_y="textures/cubemaps/yokohama/negy.jpg",
            neg_z="textures/cubemaps/yokohama/negz.jpg",
            pos_x="textures/cubemaps/yokohama/negx.jpg",
            pos_y="textures/cubemaps/yokohama/posy.jpg",
            pos_z="textures/cubemaps/yokohama/posz.jpg",
            flip_x=True,
        )
        self.prog = self.load_program("programs/cubemap.glsl")

    def on_render(self, time, frame_time):
        self.ctx.enable_only(moderngl.CULL_FACE)
        self.ctx.front_face = "cw"

        cam = self.camera.matrix
        # Purge camera translation
        cam[3][0] = 0
        cam[3][1] = 0
        cam[3][2] = 0

        self.texture.use(location=0)
        self.prog["m_proj"].write(self.camera.projection.matrix)
        self.prog["m_camera"].write(cam)

        self.cube.render(self.prog)


if __name__ == "__main__":
    Cubemap.run()



"""
Registry general data files
"""

from typing import Any

from moderngl_window.meta import DataDescription, ResourceDescription
from moderngl_window.resources.base import BaseRegistry


class DataFiles(BaseRegistry):
    """Registry for requested data files"""

    settings_attr = "DATA_LOADERS"
    meta: DataDescription

    def load(self, meta: ResourceDescription) -> Any:
        """Load data file with the configured loaders.

        Args:
            meta (:py:class:`~moderngl_window.meta.data.DataDescription`): the resource description
        Returns:
            Any: The loaded resource
        """
        return super().load(meta)


data = DataFiles()
